@@ -25,27 +25,39 @@ namespace gfx {
 struct RecordingSourceSurfaceUserData {
   void* refPtr;
   RefPtr<DrawEventRecorderPrivate> recorder;
+
+  // The optimized surface holds a reference to our surface, for GetDataSurface
+  // calls, so we must hold a weak reference to avoid circular dependency.
+  ThreadSafeWeakPtr<SourceSurface> optimizedSurface;
 };
 
 static void RecordingSourceSurfaceUserDataFunc(void* aUserData) {
   RecordingSourceSurfaceUserData* userData =
       static_cast<RecordingSourceSurfaceUserData*>(aUserData);
 
-  userData->recorder->RecordSourceSurfaceDestruction(
-      static_cast<SourceSurface*>(userData->refPtr));
-
-  delete userData;
-}
-
-static void EnsureSurfaceStoredRecording(DrawEventRecorderPrivate* aRecorder,
-                                         SourceSurface* aSurface,
-                                         const char* reason) {
-  if (aRecorder->HasStoredObject(aSurface)) {
+  if (NS_IsMainThread()) {
+    userData->recorder->RecordSourceSurfaceDestruction(userData->refPtr);
+    delete userData;
     return;
   }
 
+  userData->recorder->AddPendingDeletion([userData]() -> void {
+    userData->recorder->RecordSourceSurfaceDestruction(userData->refPtr);
+    delete userData;
+  });
+}
+
+static bool EnsureSurfaceStoredRecording(DrawEventRecorderPrivate* aRecorder,
+                                         SourceSurface* aSurface,
+                                         const char* reason) {
+  // It's important that TryAddStoredObject is called first because that will
+  // run any pending processing required by recorded objects that have been
+  // deleted off the main thread.
+  if (!aRecorder->TryAddStoredObject(aSurface)) {
+    // Surface is already stored.
+    return false;
+  }
   aRecorder->StoreSourceSurfaceRecording(aSurface, reason);
-  aRecorder->AddStoredObject(aSurface);
   aRecorder->AddSourceSurface(aSurface);
 
   RecordingSourceSurfaceUserData* userData = new RecordingSourceSurfaceUserData;
@@ -53,6 +65,7 @@ static void EnsureSurfaceStoredRecording(DrawEventRecorderPrivate* aRecorder,
   userData->recorder = aRecorder;
   aSurface->AddUserData(reinterpret_cast<UserDataKey*>(aRecorder), userData,
                         &RecordingSourceSurfaceUserDataFunc);
+  return true;
 }
 
 class SourceSurfaceRecording : public SourceSurface {
@@ -198,6 +211,15 @@ DrawTargetRecording::~DrawTargetRecording() {
   mRecorder->RecordEvent(RecordedDrawTargetDestruction(ReferencePtr(this)));
 }
 
+void DrawTargetRecording::Link(const char* aDestination, const Rect& aRect) {
+  mRecorder->RecordEvent(RecordedLink(this, aDestination, aRect));
+}
+
+void DrawTargetRecording::Destination(const char* aDestination,
+                                      const Point& aPoint) {
+  mRecorder->RecordEvent(RecordedDestination(this, aDestination, aPoint));
+}
+
 void DrawTargetRecording::FillRect(const Rect& aRect, const Pattern& aPattern,
                                    const DrawOptions& aOptions) {
   EnsurePatternDependenciesStored(aPattern);
@@ -226,6 +248,21 @@ void DrawTargetRecording::StrokeLine(const Point& aBegin, const Point& aEnd,
 
 void DrawTargetRecording::Fill(const Path* aPath, const Pattern& aPattern,
                                const DrawOptions& aOptions) {
+  if (!aPath) {
+    return;
+  }
+
+  if (aPath->GetBackendType() == BackendType::RECORDING) {
+    const PathRecording* path = static_cast<const PathRecording*>(aPath);
+    auto circle = path->AsCircle();
+    if (circle) {
+      EnsurePatternDependenciesStored(aPattern);
+      mRecorder->RecordEvent(
+          RecordedFillCircle(this, circle.value(), aPattern, aOptions));
+      return;
+    }
+  }
+
   RefPtr<PathRecording> pathRecording = EnsurePathStored(aPath);
   EnsurePatternDependenciesStored(aPattern);
 
@@ -253,6 +290,10 @@ void DrawTargetRecording::FillGlyphs(ScaledFont* aFont,
                                      const GlyphBuffer& aBuffer,
                                      const Pattern& aPattern,
                                      const DrawOptions& aOptions) {
+  if (!aFont) {
+    return;
+  }
+
   EnsurePatternDependenciesStored(aPattern);
 
   UserDataKey* userDataKey = reinterpret_cast<UserDataKey*>(mRecorder.get());
@@ -311,6 +352,10 @@ void DrawTargetRecording::Mask(const Pattern& aSource, const Pattern& aMask,
 void DrawTargetRecording::MaskSurface(const Pattern& aSource,
                                       SourceSurface* aMask, Point aOffset,
                                       const DrawOptions& aOptions) {
+  if (!aMask) {
+    return;
+  }
+
   EnsurePatternDependenciesStored(aSource);
   EnsureSurfaceStoredRecording(mRecorder, aMask, "MaskSurface");
 
@@ -321,6 +366,26 @@ void DrawTargetRecording::MaskSurface(const Pattern& aSource,
 void DrawTargetRecording::Stroke(const Path* aPath, const Pattern& aPattern,
                                  const StrokeOptions& aStrokeOptions,
                                  const DrawOptions& aOptions) {
+  if (aPath->GetBackendType() == BackendType::RECORDING) {
+    const PathRecording* path = static_cast<const PathRecording*>(aPath);
+    auto circle = path->AsCircle();
+    if (circle && circle->closed) {
+      EnsurePatternDependenciesStored(aPattern);
+      mRecorder->RecordEvent(RecordedStrokeCircle(
+          this, circle.value(), aPattern, aStrokeOptions, aOptions));
+      return;
+    }
+
+    auto line = path->AsLine();
+    if (line) {
+      EnsurePatternDependenciesStored(aPattern);
+      mRecorder->RecordEvent(RecordedStrokeLine(this, line->origin,
+                                                line->destination, aPattern,
+                                                aStrokeOptions, aOptions));
+      return;
+    }
+  }
+
   RefPtr<PathRecording> pathRecording = EnsurePathStored(aPath);
   EnsurePatternDependenciesStored(aPattern);
 
@@ -360,32 +425,43 @@ void DrawTargetRecording::DrawSurface(SourceSurface* aSurface,
                                       const Rect& aDest, const Rect& aSource,
                                       const DrawSurfaceOptions& aSurfOptions,
                                       const DrawOptions& aOptions) {
+  if (!aSurface) {
+    return;
+  }
+
   EnsureSurfaceStoredRecording(mRecorder, aSurface, "DrawSurface");
 
   mRecorder->RecordEvent(RecordedDrawSurface(this, aSurface, aDest, aSource,
                                              aSurfOptions, aOptions));
 }
 
-void DrawTargetRecording::DrawDependentSurface(
-    uint64_t aId, const Rect& aDest, const DrawSurfaceOptions& aSurfOptions,
-    const DrawOptions& aOptions) {
+void DrawTargetRecording::DrawDependentSurface(uint64_t aId,
+                                               const Rect& aDest) {
   mRecorder->AddDependentSurface(aId);
-  mRecorder->RecordEvent(
-      RecordedDrawDependentSurface(this, aId, aDest, aSurfOptions, aOptions));
+  mRecorder->RecordEvent(RecordedDrawDependentSurface(this, aId, aDest));
 }
 
-void DrawTargetRecording::DrawSurfaceWithShadow(
-    SourceSurface* aSurface, const Point& aDest, const DeviceColor& aColor,
-    const Point& aOffset, Float aSigma, CompositionOp aOp) {
+void DrawTargetRecording::DrawSurfaceWithShadow(SourceSurface* aSurface,
+                                                const Point& aDest,
+                                                const ShadowOptions& aShadow,
+                                                CompositionOp aOp) {
+  if (!aSurface) {
+    return;
+  }
+
   EnsureSurfaceStoredRecording(mRecorder, aSurface, "DrawSurfaceWithShadow");
 
-  mRecorder->RecordEvent(RecordedDrawSurfaceWithShadow(
-      this, aSurface, aDest, aColor, aOffset, aSigma, aOp));
+  mRecorder->RecordEvent(
+      RecordedDrawSurfaceWithShadow(this, aSurface, aDest, aShadow, aOp));
 }
 
 void DrawTargetRecording::DrawFilter(FilterNode* aNode, const Rect& aSourceRect,
                                      const Point& aDestPoint,
                                      const DrawOptions& aOptions) {
+  if (!aNode) {
+    return;
+  }
+
   MOZ_ASSERT(mRecorder->HasStoredObject(aNode));
 
   mRecorder->RecordEvent(
@@ -408,6 +484,10 @@ void DrawTargetRecording::ClearRect(const Rect& aRect) {
 void DrawTargetRecording::CopySurface(SourceSurface* aSurface,
                                       const IntRect& aSourceRect,
                                       const IntPoint& aDestination) {
+  if (!aSurface) {
+    return;
+  }
+
   EnsureSurfaceStoredRecording(mRecorder, aSurface, "CopySurface");
 
   mRecorder->RecordEvent(
@@ -415,6 +495,20 @@ void DrawTargetRecording::CopySurface(SourceSurface* aSurface,
 }
 
 void DrawTargetRecording::PushClip(const Path* aPath) {
+  if (!aPath) {
+    return;
+  }
+
+  // The canvas doesn't have a clipRect API so we always end up in the generic
+  // path. The D2D backend doesn't have a good way of specializing rectangular
+  // clips so we take advantage of the fact that aPath is usually backed by a
+  // SkiaPath which implements AsRect() and specialize it here.
+  auto rect = aPath->AsRect();
+  if (rect.isSome()) {
+    PushClipRect(rect.value());
+    return;
+  }
+
   RefPtr<PathRecording> pathRecording = EnsurePathStored(aPath);
 
   mRecorder->RecordEvent(RecordedPushClip(this, pathRecording));
@@ -477,19 +571,37 @@ DrawTargetRecording::CreateSourceSurfaceFromData(unsigned char* aData,
 
 already_AddRefed<SourceSurface> DrawTargetRecording::OptimizeSourceSurface(
     SourceSurface* aSurface) const {
-  if (aSurface->GetType() == SurfaceType::RECORDING &&
-      static_cast<SourceSurfaceRecording*>(aSurface)->mRecorder == mRecorder) {
-    // aSurface is already optimized for our recorder.
-    return do_AddRef(aSurface);
-  }
+  // See if we have a previously optimized surface available. We have to do this
+  // check before the SurfaceType::RECORDING below, because aSurface might be a
+  // SurfaceType::RECORDING from another recorder we have previously optimized.
+  auto* userData = static_cast<RecordingSourceSurfaceUserData*>(
+      aSurface->GetUserData(reinterpret_cast<UserDataKey*>(mRecorder.get())));
+  if (userData) {
+    RefPtr<SourceSurface> strongRef(userData->optimizedSurface);
+    if (strongRef) {
+      return do_AddRef(strongRef);
+    }
+  } else {
+    if (!EnsureSurfaceStoredRecording(mRecorder, aSurface,
+                                      "OptimizeSourceSurface")) {
+      // Surface was already stored, but doesn't have UserData so must be one
+      // of our recording surfaces.
+      MOZ_ASSERT(aSurface->GetType() == SurfaceType::RECORDING);
+      return do_AddRef(aSurface);
+    }
 
-  EnsureSurfaceStoredRecording(mRecorder, aSurface, "OptimizeSourceSurface");
+    userData = static_cast<RecordingSourceSurfaceUserData*>(
+        aSurface->GetUserData(reinterpret_cast<UserDataKey*>(mRecorder.get())));
+    MOZ_ASSERT(userData,
+               "User data should always have been set by "
+               "EnsureSurfaceStoredRecording.");
+  }
 
   RefPtr<SourceSurface> retSurf = new SourceSurfaceRecording(
       aSurface->GetSize(), aSurface->GetFormat(), mRecorder, aSurface);
-
   mRecorder->RecordEvent(
       RecordedOptimizeSourceSurface(aSurface, this, retSurf));
+  userData->optimizedSurface = retSurf;
 
   return retSurf.forget();
 }
@@ -586,8 +698,8 @@ DrawTargetRecording::CreateSimilarDrawTargetForFilter(
 
 already_AddRefed<PathBuilder> DrawTargetRecording::CreatePathBuilder(
     FillRule aFillRule) const {
-  RefPtr<PathBuilder> builder = mFinalDT->CreatePathBuilder(aFillRule);
-  return MakeAndAddRef<PathBuilderRecording>(builder, aFillRule);
+  return MakeAndAddRef<PathBuilderRecording>(mFinalDT->GetBackendType(),
+                                             aFillRule);
 }
 
 already_AddRefed<GradientStops> DrawTargetRecording::CreateGradientStops(
@@ -611,21 +723,24 @@ already_AddRefed<PathRecording> DrawTargetRecording::EnsurePathStored(
   if (aPath->GetBackendType() == BackendType::RECORDING) {
     pathRecording =
         const_cast<PathRecording*>(static_cast<const PathRecording*>(aPath));
-    if (mRecorder->HasStoredObject(aPath)) {
+    if (!mRecorder->TryAddStoredObject(pathRecording)) {
+      // Path is already stored.
       return pathRecording.forget();
     }
   } else {
     MOZ_ASSERT(!mRecorder->HasStoredObject(aPath));
     FillRule fillRule = aPath->GetFillRule();
-    RefPtr<PathBuilder> builder = mFinalDT->CreatePathBuilder(fillRule);
     RefPtr<PathBuilderRecording> builderRecording =
-        new PathBuilderRecording(builder, fillRule);
+        new PathBuilderRecording(mFinalDT->GetBackendType(), fillRule);
     aPath->StreamToSink(builderRecording);
     pathRecording = builderRecording->Finish().downcast<PathRecording>();
+    mRecorder->AddStoredObject(pathRecording);
   }
 
+  // It's important that AddStoredObject or TryAddStoredObject is called before
+  // this because that will run any pending processing required by recorded
+  // objects that have been deleted off the main thread.
   mRecorder->RecordEvent(RecordedPathCreation(pathRecording.get()));
-  mRecorder->AddStoredObject(pathRecording);
   pathRecording->mStoredRecorders.push_back(mRecorder);
 
   return pathRecording.forget();
