@@ -24,12 +24,13 @@
 #include "nsDebug.h"           // for NS_ASSERTION
 #include "nsISupportsImpl.h"   // for Image::Release, etc
 #include "nsTArray.h"          // for nsTArray
+#include "nsThreadUtils.h"     // for NS_IsMainThread
 #include "mozilla/Atomics.h"
 #include "mozilla/gfx/2D.h"
-#include "nsDataHashtable.h"
 #include "mozilla/EnumeratedArray.h"
 #include "mozilla/UniquePtr.h"
 #include "MediaInfo.h"
+#include "nsTHashMap.h"
 
 #ifdef XP_WIN
 struct ID3D10Texture2D;
@@ -48,14 +49,15 @@ class ImageCompositeNotification;
 class ImageContainer;
 class ImageContainerChild;
 class SharedPlanarYCbCrImage;
-class SharedSurfacesAnimation;
 class SurfaceDescriptor;
 class PlanarYCbCrImage;
 class TextureClient;
 class TextureClientRecycleAllocator;
 class KnowsCompositor;
 class NVImage;
+class MemoryOrShmem;
 #ifdef XP_WIN
+class D3D11RecycleAllocator;
 class D3D11YCbCrRecycleAllocator;
 #endif
 #ifdef XP_MACOSX
@@ -77,7 +79,7 @@ class SharedRGBImage;
 class SurfaceTextureImage;
 #elif defined(XP_MACOSX)
 class MacIOSurfaceImage;
-#elif MOZ_WAYLAND
+#elif MOZ_WIDGET_GTK
 class DMABUFSurfaceImage;
 #endif
 
@@ -108,6 +110,9 @@ class Image {
     return gfx::IntRect(GetOrigin().x, GetOrigin().y, GetSize().width,
                         GetSize().height);
   }
+  virtual gfx::ColorDepth GetColorDepth() const {
+    return gfx::ColorDepth::COLOR_8;
+  }
 
   ImageBackendData* GetBackendData(LayersBackend aBackend) {
     return mBackendData[aBackend].get();
@@ -118,6 +123,9 @@ class Image {
 
   int32_t GetSerial() const { return mSerial; }
 
+  bool IsDRM() const { return mIsDRM; }
+  void SetIsDRM(bool aIsDRM) { mIsDRM = aIsDRM; }
+
   virtual already_AddRefed<gfx::SourceSurface> GetAsSourceSurface() = 0;
 
   virtual bool IsValid() const { return true; }
@@ -139,7 +147,7 @@ class Image {
   virtual MacIOSurfaceImage* AsMacIOSurfaceImage() { return nullptr; }
 #endif
   virtual PlanarYCbCrImage* AsPlanarYCbCrImage() { return nullptr; }
-#ifdef MOZ_WAYLAND
+#ifdef MOZ_WIDGET_GTK
   virtual DMABUFSurfaceImage* AsDMABUFSurfaceImage() { return nullptr; }
 #endif
 
@@ -152,7 +160,10 @@ class Image {
       TextureClient* tcOverride = nullptr);
 
   Image(void* aImplData, ImageFormat aFormat)
-      : mImplData(aImplData), mSerial(++sSerialCounter), mFormat(aFormat) {}
+      : mImplData(aImplData),
+        mSerial(++sSerialCounter),
+        mFormat(aFormat),
+        mIsDRM(false) {}
 
   // Protected destructor, to discourage deletion outside of Release():
   virtual ~Image() = default;
@@ -165,6 +176,7 @@ class Image {
   void* mImplData;
   int32_t mSerial;
   ImageFormat mFormat;
+  bool mIsDRM;
 
   static mozilla::Atomic<int32_t> sSerialCounter;
 };
@@ -201,9 +213,10 @@ class BufferRecycleBin final {
 
   // We should probably do something to prune this list on a timer so we don't
   // eat excess memory while video is paused...
-  nsTArray<mozilla::UniquePtr<uint8_t[]>> mRecycledBuffers;
+  nsTArray<mozilla::UniquePtr<uint8_t[]>> mRecycledBuffers
+      MOZ_GUARDED_BY(mLock);
   // This is only valid if mRecycledBuffers is non-empty
-  uint32_t mRecycledBufferSize;
+  uint32_t mRecycledBufferSize MOZ_GUARDED_BY(mLock);
 };
 
 /**
@@ -252,7 +265,7 @@ class ImageContainerListener final {
   ~ImageContainerListener();
 
   Mutex mLock;
-  ImageContainer* mImageContainer;
+  ImageContainer* mImageContainer MOZ_GUARDED_BY(mLock);
 };
 
 /**
@@ -281,7 +294,6 @@ class ImageContainer final : public SupportsThreadSafeWeakPtr<ImageContainer> {
   friend class ImageContainerChild;
 
  public:
-  MOZ_DECLARE_THREADSAFEWEAKREFERENCE_TYPENAME(ImageContainer)
   MOZ_DECLARE_REFCOUNTED_TYPENAME(ImageContainer)
 
   enum Mode { SYNCHRONOUS = 0x0, ASYNCHRONOUS = 0x01 };
@@ -349,9 +361,11 @@ class ImageContainer final : public SupportsThreadSafeWeakPtr<ImageContainer> {
   void SetCurrentImages(const nsTArray<NonOwningImage>& aImages);
 
   /**
-   * Clear all images. Let ImageClient release all TextureClients.
+   * Clear all images. Let ImageClient release all TextureClients. Because we
+   * may release the lock after acquiring it in this method, it cannot be called
+   * with the lock held.
    */
-  void ClearAllImages();
+  void ClearAllImages() MOZ_EXCLUDES(mRecursiveMutex);
 
   /**
    * Clear any resources that are not immediately necessary. This may be called
@@ -443,36 +457,58 @@ class ImageContainer final : public SupportsThreadSafeWeakPtr<ImageContainer> {
    * Can be called on any thread. This method takes mRecursiveMutex
    * when accessing thread-shared state.
    */
-  void SetScaleHint(const gfx::IntSize& aScaleHint) { mScaleHint = aScaleHint; }
+  void SetScaleHint(const gfx::IntSize& aScaleHint) {
+    RecursiveMutexAutoLock lock(mRecursiveMutex);
+    mScaleHint = aScaleHint;
+  }
 
-  const gfx::IntSize& GetScaleHint() const { return mScaleHint; }
+  const gfx::IntSize GetScaleHint() const {
+    RecursiveMutexAutoLock lock(mRecursiveMutex);
+    return mScaleHint;
+  }
 
   void SetTransformHint(const gfx::Matrix& aTransformHint) {
+    RecursiveMutexAutoLock lock(mRecursiveMutex);
     mTransformHint = aTransformHint;
   }
 
-  const gfx::Matrix& GetTransformHint() const { return mTransformHint; }
+  const gfx::Matrix GetTransformHint() const {
+    RecursiveMutexAutoLock lock(mRecursiveMutex);
+    return mTransformHint;
+  }
 
-  void SetRotation(VideoInfo::Rotation aRotation) { mRotation = aRotation; }
+  void SetRotation(VideoInfo::Rotation aRotation) {
+    MOZ_ASSERT(NS_IsMainThread());
+    mRotation = aRotation;
+  }
 
-  VideoInfo::Rotation GetRotation() const { return mRotation; }
+  VideoInfo::Rotation GetRotation() const {
+    MOZ_ASSERT(NS_IsMainThread());
+    return mRotation;
+  }
 
   void SetImageFactory(ImageFactory* aFactory) {
     RecursiveMutexAutoLock lock(mRecursiveMutex);
     mImageFactory = aFactory ? aFactory : new ImageFactory();
   }
 
-  ImageFactory* GetImageFactory() const { return mImageFactory; }
+  already_AddRefed<ImageFactory> GetImageFactory() const {
+    RecursiveMutexAutoLock lock(mRecursiveMutex);
+    return do_AddRef(mImageFactory);
+  }
 
   void EnsureRecycleAllocatorForRDD(KnowsCompositor* aKnowsCompositor);
 
 #ifdef XP_WIN
-  D3D11YCbCrRecycleAllocator* GetD3D11YCbCrRecycleAllocator(
+  already_AddRefed<D3D11RecycleAllocator> GetD3D11RecycleAllocator(
+      KnowsCompositor* aKnowsCompositor, gfx::SurfaceFormat aPreferredFormat);
+  already_AddRefed<D3D11YCbCrRecycleAllocator> GetD3D11YCbCrRecycleAllocator(
       KnowsCompositor* aKnowsCompositor);
 #endif
 
 #ifdef XP_MACOSX
-  MacIOSurfaceRecycleAllocator* GetMacIOSurfaceRecycleAllocator();
+  already_AddRefed<MacIOSurfaceRecycleAllocator>
+  GetMacIOSurfaceRecycleAllocator();
 #endif
 
   /**
@@ -510,9 +546,7 @@ class ImageContainer final : public SupportsThreadSafeWeakPtr<ImageContainer> {
   void NotifyComposite(const ImageCompositeNotification& aNotification);
   void NotifyDropped(uint32_t aDropped);
 
-  ImageContainerListener* GetImageContainerListener() {
-    return mNotifyCompositeListener;
-  }
+  already_AddRefed<ImageContainerListener> GetImageContainerListener() const;
 
   /**
    * Get the ImageClient associated with this container. Returns only after
@@ -528,12 +562,6 @@ class ImageContainer final : public SupportsThreadSafeWeakPtr<ImageContainer> {
 
   void DropImageClient();
 
-  SharedSurfacesAnimation* GetSharedSurfacesAnimation() const {
-    return mSharedAnimation;
-  }
-
-  SharedSurfacesAnimation* EnsureSharedSurfacesAnimation();
-
  private:
   typedef mozilla::RecursiveMutex RecursiveMutex;
 
@@ -545,33 +573,44 @@ class ImageContainer final : public SupportsThreadSafeWeakPtr<ImageContainer> {
   // calling this function!
   void EnsureActiveImage();
 
-  void EnsureImageClient();
+  void EnsureImageClient() MOZ_REQUIRES(mRecursiveMutex);
+
+  bool HasImageClient() const {
+    RecursiveMutexAutoLock lock(mRecursiveMutex);
+    return !!mImageClient;
+  }
 
   // RecursiveMutex to protect thread safe access to the "current
   // image", and any other state which is shared between threads.
-  RecursiveMutex mRecursiveMutex;
+  mutable RecursiveMutex mRecursiveMutex;
 
-  RefPtr<TextureClientRecycleAllocator> mRecycleAllocator;
+  RefPtr<TextureClientRecycleAllocator> mRecycleAllocator
+      MOZ_GUARDED_BY(mRecursiveMutex);
 
 #ifdef XP_WIN
-  RefPtr<D3D11YCbCrRecycleAllocator> mD3D11YCbCrRecycleAllocator;
+  RefPtr<D3D11RecycleAllocator> mD3D11RecycleAllocator
+      MOZ_GUARDED_BY(mRecursiveMutex);
+
+  RefPtr<D3D11YCbCrRecycleAllocator> mD3D11YCbCrRecycleAllocator
+      MOZ_GUARDED_BY(mRecursiveMutex);
 #endif
 #ifdef XP_MACOSX
-  RefPtr<MacIOSurfaceRecycleAllocator> mMacIOSurfaceRecycleAllocator;
+  RefPtr<MacIOSurfaceRecycleAllocator> mMacIOSurfaceRecycleAllocator
+      MOZ_GUARDED_BY(mRecursiveMutex);
 #endif
 
-  nsTArray<OwningImage> mCurrentImages;
+  nsTArray<OwningImage> mCurrentImages MOZ_GUARDED_BY(mRecursiveMutex);
 
   // Updates every time mActiveImage changes
-  uint32_t mGenerationCounter;
+  uint32_t mGenerationCounter MOZ_GUARDED_BY(mRecursiveMutex);
 
   // Number of contained images that have been painted at least once.  It's up
   // to the ImageContainer implementation to ensure accesses to this are
   // threadsafe.
-  uint32_t mPaintCount;
+  uint32_t mPaintCount MOZ_GUARDED_BY(mRecursiveMutex);
 
   // See GetPaintDelay. Accessed only with mRecursiveMutex held.
-  TimeDuration mPaintDelay;
+  TimeDuration mPaintDelay MOZ_GUARDED_BY(mRecursiveMutex);
 
   // See GetDroppedImageCount.
   mozilla::Atomic<uint32_t> mDroppedImageCount;
@@ -579,15 +618,16 @@ class ImageContainer final : public SupportsThreadSafeWeakPtr<ImageContainer> {
   // This is the image factory used by this container, layer managers using
   // this container can set an alternative image factory that will be used to
   // create images for this container.
-  RefPtr<ImageFactory> mImageFactory;
+  RefPtr<ImageFactory> mImageFactory MOZ_GUARDED_BY(mRecursiveMutex);
 
-  gfx::IntSize mScaleHint;
+  gfx::IntSize mScaleHint MOZ_GUARDED_BY(mRecursiveMutex);
 
-  gfx::Matrix mTransformHint;
+  gfx::Matrix mTransformHint MOZ_GUARDED_BY(mRecursiveMutex);
 
+  // Main thread only.
   VideoInfo::Rotation mRotation = VideoInfo::Rotation::kDegree_0;
 
-  RefPtr<BufferRecycleBin> mRecycleBin;
+  RefPtr<BufferRecycleBin> mRecycleBin MOZ_GUARDED_BY(mRecursiveMutex);
 
   // This member points to an ImageClient if this ImageContainer was
   // sucessfully created with ENABLE_ASYNC, or points to null otherwise.
@@ -596,15 +636,13 @@ class ImageContainer final : public SupportsThreadSafeWeakPtr<ImageContainer> {
   // In this case the ImageContainer is perfectly usable, but it will forward
   // frames to the compositor through transactions in the main thread rather
   // than asynchronusly using the ImageBridge IPDL protocol.
-  RefPtr<ImageClient> mImageClient;
+  RefPtr<ImageClient> mImageClient MOZ_GUARDED_BY(mRecursiveMutex);
 
-  RefPtr<SharedSurfacesAnimation> mSharedAnimation;
-
-  bool mIsAsync;
-  CompositableHandle mAsyncContainerHandle;
+  const bool mIsAsync;
+  CompositableHandle mAsyncContainerHandle MOZ_GUARDED_BY(mRecursiveMutex);
 
   // ProducerID for last current image(s)
-  ProducerID mCurrentProducerID;
+  ProducerID mCurrentProducerID MOZ_GUARDED_BY(mRecursiveMutex);
 
   RefPtr<ImageContainerListener> mNotifyCompositeListener;
 
@@ -642,41 +680,58 @@ class AutoLockImage {
   AutoTArray<ImageContainer::OwningImage, 4> mImages;
 };
 
+// This type is currently only used for AVIF and WebCodecs therefore makes some
+// specific assumptions (e.g., Alpha's bpc and stride is equal to Y's one)
+struct PlanarAlphaData {
+  uint8_t* mChannel = nullptr;
+  gfx::IntSize mSize = gfx::IntSize(0, 0);
+  gfx::ColorDepth mDepth = gfx::ColorDepth::COLOR_8;
+  bool mPremultiplied = false;
+};
 struct PlanarYCbCrData {
   // Luminance buffer
   uint8_t* mYChannel = nullptr;
   int32_t mYStride = 0;
-  gfx::IntSize mYSize = gfx::IntSize(0, 0);
   int32_t mYSkip = 0;
   // Chroma buffers
   uint8_t* mCbChannel = nullptr;
   uint8_t* mCrChannel = nullptr;
   int32_t mCbCrStride = 0;
-  gfx::IntSize mCbCrSize = gfx::IntSize(0, 0);
   int32_t mCbSkip = 0;
   int32_t mCrSkip = 0;
+  // Alpha buffer and its metadata
+  Maybe<PlanarAlphaData> mAlpha = Nothing();
   // Picture region
-  uint32_t mPicX = 0;
-  uint32_t mPicY = 0;
-  gfx::IntSize mPicSize = gfx::IntSize(0, 0);
+  gfx::IntRect mPictureRect = gfx::IntRect(0, 0, 0, 0);
   StereoMode mStereoMode = StereoMode::MONO;
   gfx::ColorDepth mColorDepth = gfx::ColorDepth::COLOR_8;
-  gfx::YUVColorSpace mYUVColorSpace = gfx::YUVColorSpace::UNKNOWN;
+  gfx::YUVColorSpace mYUVColorSpace = gfx::YUVColorSpace::Default;
+  gfx::ColorSpace2 mColorPrimaries = gfx::ColorSpace2::UNKNOWN;
+  gfx::TransferFunction mTransferFunction = gfx::TransferFunction::BT709;
   gfx::ColorRange mColorRange = gfx::ColorRange::LIMITED;
+  gfx::ChromaSubsampling mChromaSubsampling = gfx::ChromaSubsampling::FULL;
 
-  gfx::IntRect GetPictureRect() const {
-    return gfx::IntRect(mPicX, mPicY, mPicSize.width, mPicSize.height);
+  // The cropped picture size of the Y channel.
+  gfx::IntSize YPictureSize() const { return mPictureRect.Size(); }
+
+  // The cropped picture size of the Cb/Cr channels.
+  gfx::IntSize CbCrPictureSize() const {
+    return mCbCrStride > 0 ? gfx::ChromaSize(YPictureSize(), mChromaSubsampling)
+                           : gfx::IntSize(0, 0);
   }
-};
 
-// This type is currently only used for AVIF and therefore makes some
-// AVIF-specific assumptions (e.g., Alpha's bpc and stride is equal to Y's one)
-struct PlanarYCbCrAData : PlanarYCbCrData {
-  uint8_t* mAlphaChannel = nullptr;
-  gfx::IntSize mAlphaSize = gfx::IntSize(0, 0);
-  bool mPremultipliedAlpha = false;
+  // The total uncropped size of data in the Y channel.
+  gfx::IntSize YDataSize() const {
+    return gfx::IntSize(mPictureRect.XMost(), mPictureRect.YMost());
+  }
 
-  bool hasAlpha() { return mAlphaChannel; }
+  // The total uncropped size of data in the Cb/Cr channels.
+  gfx::IntSize CbCrDataSize() const {
+    return mCbCrStride > 0 ? gfx::ChromaSize(YDataSize(), mChromaSubsampling)
+                           : gfx::IntSize(0, 0);
+  }
+
+  static Maybe<PlanarYCbCrData> From(const SurfaceDescriptorBuffer&);
 };
 
 /****** Image subtypes for the different formats ******/
@@ -691,12 +746,9 @@ struct PlanarYCbCrAData : PlanarYCbCrData {
  * 4:2:2 - CbCr width is half that of Y. Height is the same.
  * 4:2:0 - CbCr width and height is half that of Y.
  *
- * The color format is detected based on the height/width ratios
- * defined above.
+ * mChromaSubsampling specifies which YCbCr subsampling scheme to use.
  *
- * The Image that is rendered is the picture region defined by
- * mPicX, mPicY and mPicSize. The size of the rendered image is
- * mPicSize, not mYSize or mCbCrSize.
+ * The Image that is rendered is the picture region defined by mPictureRect.
  *
  * mYSkip, mCbSkip, mCrSkip are added to support various output
  * formats from hardware decoder. They are per-pixel skips in the
@@ -706,7 +758,7 @@ struct PlanarYCbCrAData : PlanarYCbCrData {
  * the mYChannel buffer looks like:
  *
  * |<----------------------- mYStride ----------------------------->|
- * |<----------------- mYSize.width --------------->|
+ * |<----------------- YDataSize().width ---------->|
  *  0   3   6   9   12  15  18  21                639             669
  * |----------------------------------------------------------------|
  * |Y___Y___Y___Y___Y___Y___Y___Y...                |%%%%%%%%%%%%%%%|
@@ -735,11 +787,15 @@ class PlanarYCbCrImage : public Image {
   virtual bool AdoptData(const Data& aData);
 
   /**
-   * Ask this Image to not convert YUV to RGB during SetData, and make
-   * the original data available through GetData. This is optional,
-   * and not all PlanarYCbCrImages will support it.
+   * This will create an empty data buffers according to the input data's size.
    */
-  virtual void SetDelayedConversion(bool aDelayed) {}
+  virtual bool CreateEmptyBuffer(const Data& aData, const gfx::IntSize& aYSize,
+                                 const gfx::IntSize& aCbCrSize) {
+    return false;
+  }
+  bool CreateEmptyBuffer(const Data& aData) {
+    return CreateEmptyBuffer(aData, aData.YDataSize(), aData.CbCrDataSize());
+  }
 
   /**
    * Grab the original YUV data. This is optional.
@@ -768,12 +824,12 @@ class PlanarYCbCrImage : public Image {
   PlanarYCbCrImage* AsPlanarYCbCrImage() override { return this; }
 
   /**
-   * Build a SurfaceDescriptorBuffer with this image.  The provided
-   * SurfaceDescriptorBuffer must already have a valid MemoryOrShmem set
-   * with a capacity large enough to hold |GetDataSize|.
+   * Build a SurfaceDescriptorBuffer with this image.  A function to allocate
+   * a MemoryOrShmem with the given capacity must be provided.
    */
   virtual nsresult BuildSurfaceDescriptorBuffer(
-      SurfaceDescriptorBuffer& aSdBuffer);
+      SurfaceDescriptorBuffer& aSdBuffer,
+      const std::function<MemoryOrShmem(uint32_t)>& aAllocate);
 
  protected:
   already_AddRefed<gfx::SourceSurface> GetAsSourceSurface() override;
@@ -877,7 +933,7 @@ class SourceSurfaceImage final : public Image {
  private:
   gfx::IntSize mSize;
   RefPtr<gfx::SourceSurface> mSourceSurface;
-  nsDataHashtable<nsUint32HashKey, RefPtr<TextureClient>> mTextureClients;
+  nsTHashMap<uint32_t, RefPtr<TextureClient>> mTextureClients;
   TextureFlags mTextureFlags;
 };
 
