@@ -9,16 +9,15 @@
 #include <stdint.h>  // for uint8_t, uint32_t, etc
 
 #include "BufferTexture.h"
-#include "GeckoProfiler.h"
 #include "IPDLActor.h"
 #include "ImageContainer.h"  // for PlanarYCbCrData, etc
-#include "Layers.h"          // for Layer, etc
 #include "MainThreadUtils.h"
 #include "gfx2DGlue.h"
 #include "gfxPlatform.h"  // for gfxPlatform
 #include "gfxUtils.h"     // for gfxUtils::GetAsLZ4Base64Str
 #include "mozilla/Atomics.h"
 #include "mozilla/Mutex.h"
+#include "mozilla/ProfilerLabels.h"
 #include "mozilla/SchedulerGroup.h"
 #include "mozilla/StaticPrefs_gfx.h"
 #include "mozilla/StaticPrefs_layers.h"
@@ -33,8 +32,6 @@
 #include "mozilla/layers/ImageBridgeChild.h"
 #include "mozilla/layers/ImageDataSerializer.h"
 #include "mozilla/layers/PTextureChild.h"
-#include "mozilla/layers/PaintThread.h"
-#include "mozilla/layers/ShadowLayers.h"
 #include "mozilla/layers/TextureClientOGL.h"
 #include "mozilla/layers/TextureClientRecycleAllocator.h"
 #include "mozilla/layers/TextureRecorded.h"
@@ -48,17 +45,14 @@
 #  include "gfxWindowsPlatform.h"
 #  include "mozilla/gfx/DeviceManagerDx.h"
 #  include "mozilla/layers/TextureD3D11.h"
-#  include "mozilla/layers/TextureDIB.h"
 #endif
-#ifdef MOZ_X11
-#  include "GLXLibrary.h"
-#  include "mozilla/layers/TextureClientX11.h"
-#endif
-#ifdef MOZ_WAYLAND
+#ifdef MOZ_WIDGET_GTK
 #  include <gtk/gtkx.h>
-
 #  include "gfxPlatformGtk.h"
 #  include "mozilla/layers/DMABUFTextureClientOGL.h"
+#  include "mozilla/widget/DMABufLibWrapper.h"
+#endif
+#ifdef MOZ_WAYLAND
 #  include "mozilla/widget/nsWaylandDisplay.h"
 #endif
 
@@ -86,7 +80,6 @@ struct TextureDeallocParams {
   RefPtr<LayersIPCChannel> allocator;
   bool clientDeallocation;
   bool syncDeallocation;
-  bool workAroundSharedSurfaceOwnershipIssue;
 };
 
 void DeallocateTextureClient(TextureDeallocParams params);
@@ -118,7 +111,6 @@ class TextureChild final : PTextureChild {
         mTextureClient(nullptr),
         mTextureData(nullptr),
         mDestroyed(false),
-        mMainThreadOnly(false),
         mIPCOpen(false),
         mOwnsTextureData(false),
         mOwnerCalledDestroy(false),
@@ -238,7 +230,6 @@ class TextureChild final : PTextureChild {
   TextureClient* mTextureClient;
   TextureData* mTextureData;
   Atomic<bool> mDestroyed;
-  bool mMainThreadOnly;
   bool mIPCOpen;
   bool mOwnsTextureData;
   bool mOwnerCalledDestroy;
@@ -273,49 +264,25 @@ static TextureType GetTextureType(gfx::SurfaceFormat aFormat,
 
 #ifdef XP_WIN
   int32_t maxTextureSize = aKnowsCompositor->GetMaxTextureSize();
-  if ((layersBackend == LayersBackend::LAYERS_D3D11 ||
-       (layersBackend == LayersBackend::LAYERS_WR &&
-        !aKnowsCompositor->UsingSoftwareWebRender())) &&
+  if ((layersBackend == LayersBackend::LAYERS_WR &&
+       !aKnowsCompositor->UsingSoftwareWebRender()) &&
       (moz2DBackend == gfx::BackendType::DIRECT2D ||
-       moz2DBackend == gfx::BackendType::DIRECT2D1_1 ||
-       (!!(aAllocFlags & ALLOC_FOR_OUT_OF_BAND_CONTENT))) &&
+       moz2DBackend == gfx::BackendType::DIRECT2D1_1) &&
       aSize.width <= maxTextureSize && aSize.height <= maxTextureSize &&
-      !(aAllocFlags & ALLOC_UPDATE_FROM_SURFACE)) {
+      !(aAllocFlags & (ALLOC_UPDATE_FROM_SURFACE | ALLOC_DO_NOT_ACCELERATE))) {
     return TextureType::D3D11;
   }
-
-  if (layersBackend != LayersBackend::LAYERS_WR &&
-      aFormat == SurfaceFormat::B8G8R8X8 &&
-      moz2DBackend == gfx::BackendType::CAIRO && NS_IsMainThread()) {
-    return TextureType::DIB;
-  }
 #endif
 
-#ifdef MOZ_WAYLAND
-  if ((layersBackend == LayersBackend::LAYERS_OPENGL ||
-       (layersBackend == LayersBackend::LAYERS_WR &&
-        !aKnowsCompositor->UsingSoftwareWebRender())) &&
-      widget::GetDMABufDevice()->IsDMABufTexturesEnabled() &&
+#ifdef MOZ_WIDGET_GTK
+  if ((layersBackend == LayersBackend::LAYERS_WR &&
+       !aKnowsCompositor->UsingSoftwareWebRender()) &&
+      widget::DMABufDevice::IsDMABufTexturesEnabled() &&
       aFormat != SurfaceFormat::A8) {
     return TextureType::DMABUF;
   }
 #endif
 
-#ifdef MOZ_X11
-  gfxSurfaceType type =
-      gfxPlatform::GetPlatform()->ScreenReferenceSurface()->GetType();
-
-  if (layersBackend == LayersBackend::LAYERS_BASIC &&
-      moz2DBackend == gfx::BackendType::CAIRO && type == gfxSurfaceType::Xlib) {
-    return TextureType::X11;
-  }
-  if (layersBackend == LayersBackend::LAYERS_OPENGL &&
-      type == gfxSurfaceType::Xlib && aFormat != SurfaceFormat::A8 &&
-      gl::sGLXLibrary.UseTextureFromPixmap()) {
-    return TextureType::X11;
-  }
-#endif
-
 #ifdef XP_MACOSX
   if (StaticPrefs::gfx_use_iosurface_textures_AtStartup()) {
     return TextureType::MacIOSurface;
@@ -323,10 +290,6 @@ static TextureType GetTextureType(gfx::SurfaceFormat aFormat,
 #endif
 
 #ifdef MOZ_WIDGET_ANDROID
-  if (gfxVars::UseAHardwareBufferContent() &&
-      aSelector == BackendSelector::Content) {
-    return TextureType::AndroidHardwareBuffer;
-  }
   if (StaticPrefs::gfx_use_surfacetexture_textures_AtStartup()) {
     return TextureType::AndroidNativeWindow;
   }
@@ -343,11 +306,7 @@ TextureType PreferredCanvasTextureType(KnowsCompositor* aKnowsCompositor) {
 
 static bool ShouldRemoteTextureType(TextureType aTextureType,
                                     BackendSelector aSelector) {
-  if (!XRE_IsContentProcess()) {
-    return false;
-  }
-
-  if (aSelector != BackendSelector::Canvas || !gfxVars::RemoteCanvasEnabled()) {
+  if (aSelector != BackendSelector::Canvas || !gfxPlatform::UseRemoteCanvas()) {
     return false;
   }
 
@@ -381,7 +340,7 @@ TextureData* TextureData::Create(TextureForwarder* aAllocator,
     textureType = TextureType::Unknown;
   }
 
-#if defined(XP_MACOSX) || defined(MOZ_WAYLAND)
+#if defined(XP_MACOSX) || defined(MOZ_WIDGET_GTK)
   gfx::BackendType moz2DBackend = BackendTypeForBackendSelector(
       aKnowsCompositor->GetCompositorBackendType(), aSelector);
 #endif
@@ -390,26 +349,18 @@ TextureData* TextureData::Create(TextureForwarder* aAllocator,
 #ifdef XP_WIN
     case TextureType::D3D11:
       return D3D11TextureData::Create(aSize, aFormat, aAllocFlags);
-    case TextureType::DIB:
-      return DIBTextureData::Create(aSize, aFormat, aAllocator);
 #endif
 
-#ifdef MOZ_WAYLAND
+#ifdef MOZ_WIDGET_GTK
     case TextureType::DMABUF:
       return DMABUFTextureData::Create(aSize, aFormat, moz2DBackend);
 #endif
 
-#ifdef MOZ_X11
-    case TextureType::X11:
-      return X11TextureData::Create(aSize, aFormat, aTextureFlags, aAllocator);
-#endif
 #ifdef XP_MACOSX
     case TextureType::MacIOSurface:
       return MacIOSurfaceTextureData::Create(aSize, aFormat, moz2DBackend);
 #endif
 #ifdef MOZ_WIDGET_ANDROID
-    case TextureType::AndroidHardwareBuffer:
-      return AndroidHardwareBufferTextureData::Create(aSize, aFormat);
     case TextureType::AndroidNativeWindow:
       return AndroidNativeWindowTextureData::Create(aSize, aFormat);
 #endif
@@ -429,24 +380,11 @@ bool TextureData::IsRemote(KnowsCompositor* aKnowsCompositor,
 }
 
 static void DestroyTextureData(TextureData* aTextureData,
-                               LayersIPCChannel* aAllocator, bool aDeallocate,
-                               bool aMainThreadOnly) {
+                               LayersIPCChannel* aAllocator, bool aDeallocate) {
   if (!aTextureData) {
     return;
   }
 
-  if (aMainThreadOnly && !NS_IsMainThread()) {
-    RefPtr<LayersIPCChannel> allocatorRef = aAllocator;
-    SchedulerGroup::Dispatch(
-        TaskCategory::Other,
-        NS_NewRunnableFunction(
-            "layers::DestroyTextureData",
-            [aTextureData, allocatorRef, aDeallocate]() -> void {
-              DestroyTextureData(aTextureData, allocatorRef, aDeallocate, true);
-            }));
-    return;
-  }
-
   if (aDeallocate) {
     aTextureData->Deallocate(aAllocator);
   } else {
@@ -461,8 +399,7 @@ void TextureChild::ActorDestroy(ActorDestroyReason why) {
   mIPCOpen = false;
 
   if (mTextureData) {
-    DestroyTextureData(mTextureData, GetAllocator(), mOwnsTextureData,
-                       mMainThreadOnly);
+    DestroyTextureData(mTextureData, GetAllocator(), mOwnsTextureData);
     mTextureData = nullptr;
   }
 }
@@ -477,7 +414,7 @@ void TextureChild::Destroy(const TextureDeallocParams& aParams) {
 
   if (!IPCOpen()) {
     DestroyTextureData(aParams.data, aParams.allocator,
-                       aParams.clientDeallocation, mMainThreadOnly);
+                       aParams.clientDeallocation);
     return;
   }
 
@@ -530,7 +467,7 @@ void DeallocateTextureClient(TextureDeallocParams params) {
   if (ipdlThread && !ipdlThread->IsOnCurrentThread()) {
     if (params.syncDeallocation) {
       bool done = false;
-      ReentrantMonitor barrier("DeallocateTextureClient");
+      ReentrantMonitor barrier MOZ_UNANNOTATED("DeallocateTextureClient");
       ReentrantMonitorAutoEnter autoMon(barrier);
       ipdlThread->Dispatch(NewRunnableFunction(
           "DeallocateTextureClientSyncProxyRunnable",
@@ -562,13 +499,7 @@ void DeallocateTextureClient(TextureDeallocParams params) {
     // TextureClient before sharing it with the compositor. It means the data
     // cannot be owned by the TextureHost since we never created the
     // TextureHost...
-    // ..except if the lovely mWorkaroundAnnoyingSharedSurfaceOwnershipIssues
-    // member is set to true. In this case we are in a special situation where
-    // this TextureClient is in wrapped into another TextureClient which assumes
-    // it owns our data.
-    bool shouldDeallocate = !params.workAroundSharedSurfaceOwnershipIssue;
-    DestroyTextureData(params.data, params.allocator, shouldDeallocate,
-                       false);  // main-thread deallocation
+    DestroyTextureData(params.data, params.allocator, /* aDeallocate */ true);
     return;
   }
 
@@ -595,22 +526,14 @@ void TextureClient::Destroy() {
   }
 
   TextureData* data = mData;
-  if (!mWorkaroundAnnoyingSharedSurfaceLifetimeIssues) {
-    mData = nullptr;
-  }
+  mData = nullptr;
 
   if (data || actor) {
     TextureDeallocParams params;
     params.actor = actor;
     params.allocator = mAllocator;
     params.clientDeallocation = !!(mFlags & TextureFlags::DEALLOCATE_CLIENT);
-    params.workAroundSharedSurfaceOwnershipIssue =
-        mWorkaroundAnnoyingSharedSurfaceOwnershipIssues;
-    if (mWorkaroundAnnoyingSharedSurfaceLifetimeIssues) {
-      params.data = nullptr;
-    } else {
-      params.data = data;
-    }
+    params.data = data;
     // At the moment we always deallocate synchronously when deallocating on the
     // client side, but having asynchronous deallocate in some of the cases will
     // be a worthwhile optimization.
@@ -737,13 +660,6 @@ void TextureClient::Unlock() {
     if (!(mOpenMode & OpenMode::OPEN_ASYNC)) {
       if (mOpenMode & OpenMode::OPEN_WRITE) {
         mBorrowedDrawTarget->Flush();
-        if (mReadbackSink && !mData->ReadBack(mReadbackSink)) {
-          // Fallback implementation for reading back, because mData does not
-          // have a backend-specific implementation and returned false.
-          RefPtr<SourceSurface> snapshot = mBorrowedDrawTarget->Snapshot();
-          RefPtr<DataSourceSurface> dataSurf = snapshot->GetDataSurface();
-          mReadbackSink->ProcessReadback(dataSurf);
-        }
       }
 
       mBorrowedDrawTarget->DetachAllSnapshots();
@@ -879,6 +795,20 @@ gfx::DrawTarget* TextureClient::BorrowDrawTarget() {
   return mBorrowedDrawTarget;
 }
 
+void TextureClient::EndDraw() {
+  MOZ_ASSERT(mOpenMode & OpenMode::OPEN_READ_WRITE);
+
+  // Because EndDraw is used when we are not unlocking this TextureClient at the
+  // end of a transaction, we need to Flush and DetachAllSnapshots to ensure any
+  // dependents are updated.
+  mBorrowedDrawTarget->Flush();
+  mBorrowedDrawTarget->DetachAllSnapshots();
+  MOZ_ASSERT(mBorrowedDrawTarget->refCount() <= mExpectedDtRefs);
+
+  mBorrowedDrawTarget = nullptr;
+  mData->EndDraw();
+}
+
 already_AddRefed<gfx::SourceSurface> TextureClient::BorrowSnapshot() {
   MOZ_ASSERT(mIsLocked);
 
@@ -1070,13 +1000,6 @@ bool TextureClient::InitIPDLActor(CompositableForwarder* aForwarder) {
         MOZ_ASSERT_UNREACHABLE("unexpected to be called");
         return false;
       }
-      if (ShadowLayerForwarder* forwarder = aForwarder->AsLayerForwarder()) {
-        // Do the DOM labeling.
-        if (nsISerialEventTarget* target = forwarder->GetEventTarget()) {
-          forwarder->GetCompositorBridgeChild()->ReplaceEventTargetForActor(
-              mActor, target);
-        }
-      }
       mActor->mCompositableForwarder = aForwarder;
       mActor->mUsesImageBridge =
           aForwarder->GetTextureForwarder()->UsesImageBridge();
@@ -1095,20 +1018,15 @@ bool TextureClient::InitIPDLActor(CompositableForwarder* aForwarder) {
   mExternalImageId =
       aForwarder->GetTextureForwarder()->GetNextExternalImageId();
 
-  nsISerialEventTarget* target = nullptr;
-  // Get the layers id if the forwarder is a ShadowLayerForwarder.
-  if (ShadowLayerForwarder* forwarder = aForwarder->AsLayerForwarder()) {
-    target = forwarder->GetEventTarget();
-  }
-
   ReadLockDescriptor readLockDescriptor = null_t();
   if (mReadLock) {
     mReadLock->Serialize(readLockDescriptor, GetAllocator()->GetParentPid());
   }
 
   PTextureChild* actor = aForwarder->GetTextureForwarder()->CreateTexture(
-      desc, readLockDescriptor, aForwarder->GetCompositorBackendType(),
-      GetFlags(), mSerial, mExternalImageId, target);
+      desc, std::move(readLockDescriptor),
+      aForwarder->GetCompositorBackendType(), GetFlags(), mSerial,
+      mExternalImageId);
 
   if (!actor) {
     gfxCriticalNote << static_cast<int32_t>(desc.type()) << ", "
@@ -1123,7 +1041,6 @@ bool TextureClient::InitIPDLActor(CompositableForwarder* aForwarder) {
   mActor->mCompositableForwarder = aForwarder;
   mActor->mTextureForwarder = aForwarder->GetTextureForwarder();
   mActor->mTextureClient = this;
-  mActor->mMainThreadOnly = !!(mFlags & TextureFlags::DEALLOCATE_MAIN_THREAD);
 
   // If the TextureClient is already locked, we have to lock TextureChild's
   // mutex since it will be unlocked in TextureClient::Unlock.
@@ -1174,9 +1091,10 @@ bool TextureClient::InitIPDLActor(KnowsCompositor* aKnowsCompositor) {
     mReadLock->Serialize(readLockDescriptor, GetAllocator()->GetParentPid());
   }
 
-  PTextureChild* actor = fwd->CreateTexture(
-      desc, readLockDescriptor, aKnowsCompositor->GetCompositorBackendType(),
-      GetFlags(), mSerial, mExternalImageId);
+  PTextureChild* actor =
+      fwd->CreateTexture(desc, std::move(readLockDescriptor),
+                         aKnowsCompositor->GetCompositorBackendType(),
+                         GetFlags(), mSerial, mExternalImageId);
   if (!actor) {
     gfxCriticalNote << static_cast<int32_t>(desc.type()) << ", "
                     << static_cast<int32_t>(
@@ -1189,7 +1107,6 @@ bool TextureClient::InitIPDLActor(KnowsCompositor* aKnowsCompositor) {
   mActor = static_cast<TextureChild*>(actor);
   mActor->mTextureForwarder = fwd;
   mActor->mTextureClient = this;
-  mActor->mMainThreadOnly = !!(mFlags & TextureFlags::DEALLOCATE_MAIN_THREAD);
 
   // If the TextureClient is already locked, we have to lock TextureChild's
   // mutex since it will be unlocked in TextureClient::Unlock.
@@ -1207,11 +1124,6 @@ already_AddRefed<TextureClient> TextureClient::CreateForDrawing(
     KnowsCompositor* aAllocator, gfx::SurfaceFormat aFormat, gfx::IntSize aSize,
     BackendSelector aSelector, TextureFlags aTextureFlags,
     TextureAllocationFlags aAllocFlags) {
-  if (aAllocator->SupportsTextureDirectMapping() &&
-      std::max(aSize.width, aSize.height) <= aAllocator->GetMaxTextureSize()) {
-    aAllocFlags =
-        TextureAllocationFlags(aAllocFlags | ALLOC_ALLOW_DIRECT_MAPPING);
-  }
   return TextureClient::CreateForDrawing(aAllocator->GetTextureForwarder(),
                                          aFormat, aSize, aAllocator, aSelector,
                                          aTextureFlags, aAllocFlags);
@@ -1274,12 +1186,9 @@ already_AddRefed<TextureClient> TextureClient::CreateFromSurface(
 
   int32_t maxTextureSize = aAllocator->GetMaxTextureSize();
 
-  if ((layersBackend == LayersBackend::LAYERS_D3D11 ||
-       layersBackend == LayersBackend::LAYERS_WR) &&
+  if (layersBackend == LayersBackend::LAYERS_WR &&
       (moz2DBackend == gfx::BackendType::DIRECT2D ||
-       moz2DBackend == gfx::BackendType::DIRECT2D1_1 ||
-       (!!(aAllocFlags & ALLOC_FOR_OUT_OF_BAND_CONTENT) &&
-        DeviceManagerDx::Get()->GetContentDevice())) &&
+       moz2DBackend == gfx::BackendType::DIRECT2D1_1) &&
       size.width <= maxTextureSize && size.height <= maxTextureSize) {
     data = D3D11TextureData::Create(aSurface, aAllocFlags);
   }
@@ -1315,19 +1224,6 @@ already_AddRefed<TextureClient> TextureClient::CreateForRawBufferAccess(
     KnowsCompositor* aAllocator, gfx::SurfaceFormat aFormat, gfx::IntSize aSize,
     gfx::BackendType aMoz2DBackend, TextureFlags aTextureFlags,
     TextureAllocationFlags aAllocFlags) {
-  // If we exceed the max texture size for the GPU, then just fall back to no
-  // texture direct mapping. If it becomes a problem we can implement tiling
-  // logic inside DirectMapTextureSource to allow this.
-  bool supportsTextureDirectMapping =
-      aAllocator->SupportsTextureDirectMapping() &&
-      std::max(aSize.width, aSize.height) <= aAllocator->GetMaxTextureSize();
-  if (supportsTextureDirectMapping) {
-    aAllocFlags =
-        TextureAllocationFlags(aAllocFlags | ALLOC_ALLOW_DIRECT_MAPPING);
-  } else {
-    aAllocFlags =
-        TextureAllocationFlags(aAllocFlags & ~ALLOC_ALLOW_DIRECT_MAPPING);
-  }
   return CreateForRawBufferAccess(
       aAllocator->GetTextureForwarder(), aFormat, aSize, aMoz2DBackend,
       aAllocator->GetCompositorBackendType(), aTextureFlags, aAllocFlags);
@@ -1344,10 +1240,6 @@ already_AddRefed<TextureClient> TextureClient::CreateForRawBufferAccess(
     return nullptr;
   }
 
-  if (aAllocFlags & ALLOC_DISALLOW_BUFFERTEXTURECLIENT) {
-    return nullptr;
-  }
-
   if (!gfx::Factory::AllowedSurfaceSize(aSize)) {
     return nullptr;
   }
@@ -1382,7 +1274,8 @@ already_AddRefed<TextureClient> TextureClient::CreateForYCbCr(
     const gfx::IntSize& aYSize, uint32_t aYStride,
     const gfx::IntSize& aCbCrSize, uint32_t aCbCrStride, StereoMode aStereoMode,
     gfx::ColorDepth aColorDepth, gfx::YUVColorSpace aYUVColorSpace,
-    gfx::ColorRange aColorRange, TextureFlags aTextureFlags) {
+    gfx::ColorRange aColorRange, gfx::ChromaSubsampling aSubsampling,
+    TextureFlags aTextureFlags) {
   if (!aAllocator || !aAllocator->GetLayersIPCActor()->IPCOpen()) {
     return nullptr;
   }
@@ -1393,7 +1286,8 @@ already_AddRefed<TextureClient> TextureClient::CreateForYCbCr(
 
   TextureData* data = BufferTextureData::CreateForYCbCr(
       aAllocator, aDisplay, aYSize, aYStride, aCbCrSize, aCbCrStride,
-      aStereoMode, aColorDepth, aYUVColorSpace, aColorRange, aTextureFlags);
+      aStereoMode, aColorDepth, aYUVColorSpace, aColorRange, aSubsampling,
+      aTextureFlags);
   if (!data) {
     return nullptr;
   }
@@ -1419,8 +1313,6 @@ TextureClient::TextureClient(TextureData* aData, TextureFlags aFlags,
       mIsReadLocked(false),
       mUpdated(false),
       mAddedToCompositableClient(false),
-      mWorkaroundAnnoyingSharedSurfaceLifetimeIssues(false),
-      mWorkaroundAnnoyingSharedSurfaceOwnershipIssues(false),
       mFwdTransactionId(0),
       mSerial(++sSerialCounter)
 #ifdef GFX_DEBUG_TRACK_CLIENTS_IN_POOL
@@ -1489,27 +1381,6 @@ already_AddRefed<gfx::DataSourceSurface> TextureClient::GetAsSurface() {
   return data.forget();
 }
 
-void TextureClient::PrintInfo(std::stringstream& aStream, const char* aPrefix) {
-  aStream << aPrefix;
-  aStream << nsPrintfCString("TextureClient (0x%p)", this).get()
-          << " [size=" << GetSize() << "]"
-          << " [format=" << GetFormat() << "]"
-          << " [flags=" << mFlags << "]";
-
-#ifdef MOZ_DUMP_PAINTING
-  if (StaticPrefs::layers_dump_texture()) {
-    nsAutoCString pfx(aPrefix);
-    pfx += "  ";
-
-    aStream << "\n" << pfx.get() << "Surface: ";
-    RefPtr<gfx::DataSourceSurface> dSurf = GetAsSurface();
-    if (dSurf) {
-      aStream << gfxUtils::GetAsLZ4Base64Str(dSurf).get();
-    }
-  }
-#endif
-}
-
 void TextureClient::GetSurfaceDescriptorRemoteDecoder(
     SurfaceDescriptorRemoteDecoder* const aOutDesc) {
   const auto handle = GetSerial();
@@ -1597,7 +1468,8 @@ class CrossProcessSemaphoreReadLock : public TextureReadLock {
       : mSemaphore(CrossProcessSemaphore::Create("TextureReadLock", 1)),
         mShared(false) {}
   explicit CrossProcessSemaphoreReadLock(CrossProcessSemaphoreHandle aHandle)
-      : mSemaphore(CrossProcessSemaphore::Create(aHandle)), mShared(false) {}
+      : mSemaphore(CrossProcessSemaphore::Create(std::move(aHandle))),
+        mShared(false) {}
 
   bool ReadLock() override {
     if (!IsValid()) {
@@ -1630,7 +1502,7 @@ class CrossProcessSemaphoreReadLock : public TextureReadLock {
 
 // static
 already_AddRefed<TextureReadLock> TextureReadLock::Deserialize(
-    const ReadLockDescriptor& aDescriptor, ISurfaceAllocator* aAllocator) {
+    ReadLockDescriptor&& aDescriptor, ISurfaceAllocator* aAllocator) {
   switch (aDescriptor.type()) {
     case ReadLockDescriptor::TShmemSection: {
       const ShmemSection& section = aDescriptor.get_ShmemSection();
@@ -1659,7 +1531,7 @@ already_AddRefed<TextureReadLock> TextureReadLock::Deserialize(
     }
     case ReadLockDescriptor::TCrossProcessSemaphoreDescriptor: {
       return MakeAndAddRef<CrossProcessSemaphoreReadLock>(
-          aDescriptor.get_CrossProcessSemaphoreDescriptor().sem());
+          std::move(aDescriptor.get_CrossProcessSemaphoreDescriptor().sem()));
     }
     case ReadLockDescriptor::Tnull_t: {
       return nullptr;
@@ -1793,7 +1665,7 @@ bool CrossProcessSemaphoreReadLock::Serialize(ReadLockDescriptor& aOutput,
                                               base::ProcessId aOther) {
   if (!mShared && IsValid()) {
     aOutput = ReadLockDescriptor(
-        CrossProcessSemaphoreDescriptor(mSemaphore->ShareToProcess(aOther)));
+        CrossProcessSemaphoreDescriptor(mSemaphore->CloneHandle()));
     mSemaphore->CloseHandle();
     mShared = true;
     return true;
@@ -1808,17 +1680,6 @@ void TextureClient::EnableBlockingReadLock() {
   }
 }
 
-void TextureClient::AddPaintThreadRef() {
-  MOZ_ASSERT(NS_IsMainThread());
-  mPaintThreadRefs += 1;
-}
-
-void TextureClient::DropPaintThreadRef() {
-  MOZ_RELEASE_ASSERT(PaintThread::Get()->IsOnPaintWorkerThread());
-  MOZ_RELEASE_ASSERT(mPaintThreadRefs >= 1);
-  mPaintThreadRefs -= 1;
-}
-
 bool UpdateYCbCrTextureClient(TextureClient* aTexture,
                               const PlanarYCbCrData& aData) {
   MOZ_ASSERT(aTexture);
@@ -1839,17 +1700,17 @@ bool UpdateYCbCrTextureClient(TextureClient* aTexture,
       BytesPerPixel(SurfaceFormatForColorDepth(aData.mColorDepth));
   MappedYCbCrTextureData srcData;
   srcData.y.data = aData.mYChannel;
-  srcData.y.size = aData.mYSize;
+  srcData.y.size = aData.YDataSize();
   srcData.y.stride = aData.mYStride;
   srcData.y.skip = aData.mYSkip;
   srcData.y.bytesPerPixel = bytesPerPixel;
   srcData.cb.data = aData.mCbChannel;
-  srcData.cb.size = aData.mCbCrSize;
+  srcData.cb.size = aData.CbCrDataSize();
   srcData.cb.stride = aData.mCbCrStride;
   srcData.cb.skip = aData.mCbSkip;
   srcData.cb.bytesPerPixel = bytesPerPixel;
   srcData.cr.data = aData.mCrChannel;
-  srcData.cr.size = aData.mCbCrSize;
+  srcData.cr.size = aData.CbCrDataSize();
   srcData.cr.stride = aData.mCbCrStride;
   srcData.cr.skip = aData.mCrSkip;
   srcData.cr.bytesPerPixel = bytesPerPixel;
