@@ -6,14 +6,16 @@
 
 #include "CanvasTranslator.h"
 
+#include "gfxGradientCache.h"
 #include "mozilla/gfx/2D.h"
 #include "mozilla/gfx/GPUParent.h"
 #include "mozilla/gfx/Logging.h"
 #include "mozilla/ipc/Endpoint.h"
+#include "mozilla/layers/SharedSurfacesParent.h"
 #include "mozilla/layers/TextureClient.h"
 #include "mozilla/SyncRunnable.h"
 #include "mozilla/Telemetry.h"
-#include "nsTHashtable.h"
+#include "nsTHashSet.h"
 #include "RecordedCanvasEventImpl.h"
 
 #if defined(XP_WIN)
@@ -29,6 +31,9 @@ namespace layers {
 // other content processes are waiting for events to process.
 static const TimeDuration kReadEventTimeout = TimeDuration::FromMilliseconds(5);
 
+static const TimeDuration kDescriptorTimeout =
+    TimeDuration::FromMilliseconds(10000);
+
 class RingBufferReaderServices final
     : public CanvasEventRingBuffer::ReaderServices {
  public:
@@ -64,7 +69,7 @@ TextureData* CanvasTranslator::CreateTextureData(TextureType aTextureType,
   return textureData;
 }
 
-typedef nsTHashtable<nsRefPtrHashKey<CanvasTranslator>> CanvasTranslatorSet;
+typedef nsTHashSet<RefPtr<CanvasTranslator>> CanvasTranslatorSet;
 
 static CanvasTranslatorSet& CanvasTranslators() {
   MOZ_ASSERT(CanvasThreadHolder::IsInCanvasThread());
@@ -73,8 +78,8 @@ static CanvasTranslatorSet& CanvasTranslators() {
 }
 
 static void EnsureAllClosed() {
-  for (auto iter = CanvasTranslators().Iter(); !iter.Done(); iter.Next()) {
-    iter.Get()->GetKey()->Close();
+  for (const auto& key : CanvasTranslators()) {
+    key->Close();
   }
 }
 
@@ -107,9 +112,10 @@ CanvasTranslator::CanvasTranslator(
 }
 
 CanvasTranslator::~CanvasTranslator() {
-  if (mReferenceTextureData) {
-    mReferenceTextureData->Unlock();
-  }
+  // The textures need to be the last thing holding their DrawTargets, so that
+  // they can destroy them within a lock.
+  mDrawTargets.Clear();
+  mBaseDT = nullptr;
 }
 
 void CanvasTranslator::Bind(Endpoint<PCanvasParent>&& aEndpoint) {
@@ -117,20 +123,25 @@ void CanvasTranslator::Bind(Endpoint<PCanvasParent>&& aEndpoint) {
     return;
   }
 
-  CanvasTranslators().PutEntry(this);
+  CanvasTranslators().Insert(this);
 }
 
 mozilla::ipc::IPCResult CanvasTranslator::RecvInitTranslator(
     const TextureType& aTextureType,
-    const ipc::SharedMemoryBasic::Handle& aReadHandle,
-    const CrossProcessSemaphoreHandle& aReaderSem,
-    const CrossProcessSemaphoreHandle& aWriterSem) {
+    ipc::SharedMemoryBasic::Handle&& aReadHandle,
+    CrossProcessSemaphoreHandle&& aReaderSem,
+    CrossProcessSemaphoreHandle&& aWriterSem) {
+  if (mStream) {
+    return IPC_FAIL(this, "RecvInitTranslator called twice.");
+  }
+
   mTextureType = aTextureType;
 
   // We need to initialize the stream first, because it might be used to
   // communicate other failures back to the writer.
   mStream = MakeUnique<CanvasEventRingBuffer>();
-  if (!mStream->InitReader(aReadHandle, aReaderSem, aWriterSem,
+  if (!mStream->InitReader(std::move(aReadHandle), std::move(aReaderSem),
+                           std::move(aWriterSem),
                            MakeUnique<RingBufferReaderServices>(this))) {
     return IPC_FAIL(this, "Failed to initialize ring buffer reader.");
   }
@@ -206,7 +217,7 @@ void CanvasTranslator::FinishShutdown() {
   // ReleaseOnCompositorThread.
   CanvasTranslatorSet& canvasTranslators = CanvasTranslators();
   CanvasThreadHolder::ReleaseOnCompositorThread(mCanvasThreadHolder.forget());
-  canvasTranslators.RemoveEntry(this);
+  canvasTranslators.Remove(this);
 }
 
 void CanvasTranslator::Deactivate() {
@@ -222,12 +233,9 @@ void CanvasTranslator::Deactivate() {
       NewRunnableMethod("CanvasTranslator::SendDeactivate", this,
                         &CanvasTranslator::SendDeactivate));
 
-  {
-    // Unlock all of our textures.
-    gfx::AutoSerializeWithMoz2D serializeWithMoz2D(GetBackendType());
-    for (auto const& entry : mTextureDatas) {
-      entry.second->Unlock();
-    }
+  // Unlock all of our textures.
+  for (auto const& entry : mTextureDatas) {
+    entry.second->Unlock();
   }
 
   // Also notify anyone waiting for a surface descriptor. This must be done
@@ -238,7 +246,7 @@ void CanvasTranslator::Deactivate() {
 bool CanvasTranslator::TranslateRecording() {
   MOZ_ASSERT(CanvasThreadHolder::IsInCanvasWorker());
 
-  int32_t eventType = mStream->ReadNextEvent();
+  uint8_t eventType = mStream->ReadNextEvent();
   while (mStream->good()) {
     bool success = RecordedEvent::DoWithEventFromStream(
         *mStream, static_cast<RecordedEvent::EventType>(eventType),
@@ -330,7 +338,7 @@ void CanvasTranslator::Flush() {
     return;
   }
 
-  gfx::AutoSerializeWithMoz2D serializeWithMoz2D(GetBackendType());
+  gfx::AutoSerializeWithMoz2D serializeWithMoz2D(mBackendType);
   RefPtr<ID3D11DeviceContext> deviceContext;
   mDevice->GetImmediateContext(getter_AddRefs(deviceContext));
   deviceContext->Flush();
@@ -362,9 +370,13 @@ bool CanvasTranslator::CreateReferenceTexture() {
 
   mReferenceTextureData->Lock(OpenMode::OPEN_READ_WRITE);
   mBaseDT = mReferenceTextureData->BorrowDrawTarget();
-  if (mBaseDT) {
-    mBackendType = mBaseDT->GetBackendType();
+  if (!mBaseDT) {
+    // We might get a null draw target due to a device failure, just return
+    // false so that we can recover.
+    return false;
   }
+
+  mBackendType = mBaseDT->GetBackendType();
   return true;
 }
 
@@ -397,7 +409,7 @@ bool CanvasTranslator::CheckForFreshCanvasDevice(int aLineNumber) {
 
   // It is safe to wait here because only the Compositor thread waits on us and
   // the main thread doesn't wait on the compositor thread in the GPU process.
-  SyncRunnable::DispatchToThread(GetMainThreadEventTarget(), runnable,
+  SyncRunnable::DispatchToThread(GetMainThreadSerialEventTarget(), runnable,
                                  /*aForceDispatch*/ true);
 
   mDevice = gfx::DeviceManagerDx::Get()->GetCanvasDevice();
@@ -439,10 +451,6 @@ already_AddRefed<gfx::DrawTarget> CanvasTranslator::CreateDrawTarget(
     gfx::SurfaceFormat aFormat) {
   RefPtr<gfx::DrawTarget> dt;
   do {
-    // It is important that AutoSerializeWithMoz2D is called within the loop
-    // and doesn't hold during calls to CheckForFreshCanvasDevice, because that
-    // might cause a deadlock with device reset code on the main thread.
-    gfx::AutoSerializeWithMoz2D serializeWithMoz2D(GetBackendType());
     TextureData* textureData = CreateTextureData(mTextureType, aSize, aFormat);
     if (textureData) {
       MOZ_DIAGNOSTIC_ASSERT(mNextTextureId >= 0, "No texture ID set");
@@ -459,7 +467,6 @@ already_AddRefed<gfx::DrawTarget> CanvasTranslator::CreateDrawTarget(
 }
 
 void CanvasTranslator::RemoveTexture(int64_t aTextureId) {
-  gfx::AutoSerializeWithMoz2D serializeWithMoz2D(GetBackendType());
   mTextureDatas.erase(aTextureId);
 
   // It is possible that the texture from the content process has never been
@@ -487,7 +494,12 @@ UniquePtr<SurfaceDescriptor> CanvasTranslator::WaitForSurfaceDescriptor(
       return nullptr;
     }
 
-    mSurfaceDescriptorsMonitor.Wait();
+    CVStatus status = mSurfaceDescriptorsMonitor.Wait(kDescriptorTimeout);
+    if (status == CVStatus::Timeout) {
+      // If something has gone wrong and the texture has already been destroyed,
+      // it will have cleaned up its descriptor.
+      return nullptr;
+    }
   }
 
   UniquePtr<SurfaceDescriptor> descriptor = std::move(result->second);
@@ -495,6 +507,26 @@ UniquePtr<SurfaceDescriptor> CanvasTranslator::WaitForSurfaceDescriptor(
   return descriptor;
 }
 
+already_AddRefed<gfx::SourceSurface> CanvasTranslator::LookupExternalSurface(
+    uint64_t aKey) {
+  return SharedSurfacesParent::Get(wr::ToExternalImageId(aKey));
+}
+
+already_AddRefed<gfx::GradientStops> CanvasTranslator::GetOrCreateGradientStops(
+    gfx::GradientStop* aRawStops, uint32_t aNumStops,
+    gfx::ExtendMode aExtendMode) {
+  nsTArray<gfx::GradientStop> rawStopArray(aRawStops, aNumStops);
+  RefPtr<DrawTarget> drawTarget = GetReferenceDrawTarget();
+  if (!drawTarget) {
+    // We might end up with a null reference draw target due to a device
+    // failure, just return false so that we can recover.
+    return nullptr;
+  }
+
+  return gfx::gfxGradientCache::GetOrCreateGradientStops(
+      drawTarget, rawStopArray, aExtendMode);
+}
+
 gfx::DataSourceSurface* CanvasTranslator::LookupDataSurface(
     gfx::ReferencePtr aRefPtr) {
   return mDataSurfaces.GetWeak(aRefPtr);
@@ -502,7 +534,7 @@ gfx::DataSourceSurface* CanvasTranslator::LookupDataSurface(
 
 void CanvasTranslator::AddDataSurface(
     gfx::ReferencePtr aRefPtr, RefPtr<gfx::DataSourceSurface>&& aSurface) {
-  mDataSurfaces.Put(aRefPtr, std::move(aSurface));
+  mDataSurfaces.InsertOrUpdate(aRefPtr, std::move(aSurface));
 }
 
 void CanvasTranslator::RemoveDataSurface(gfx::ReferencePtr aRefPtr) {
