@@ -12,7 +12,6 @@
 #include "mozilla/StaticPrefs_gfx.h"
 #include "mozilla/StaticPrefs_layers.h"
 #include "mozilla/Telemetry.h"
-#include "mozilla/WindowsVersion.h"
 #include "mozilla/gfx/GPUParent.h"
 #include "mozilla/gfx/GPUProcessManager.h"
 #include "mozilla/gfx/GraphicsMessages.h"
@@ -21,21 +20,16 @@
 #include "mozilla/layers/CompositorBridgeChild.h"
 #include "mozilla/layers/CompositorThread.h"
 #include "mozilla/layers/DeviceAttachmentsD3D11.h"
-#include "mozilla/layers/MLGDeviceD3D11.h"
-#include "mozilla/layers/PaintThread.h"
 #include "mozilla/Preferences.h"
-#include "nsExceptionHandler.h"
 #include "nsPrintfCString.h"
 #include "nsString.h"
 
-#undef _WIN32_WINNT
-#define _WIN32_WINNT _WIN32_WINNT_WINBLUE
-#undef NTDDI_VERSION
-#define NTDDI_VERSION NTDDI_WINBLUE
+// -
 
+#include "mozilla/gfx/AllOfDcomp.h"
 #include <d3d11.h>
-#include <dcomp.h>
 #include <ddraw.h>
+#include <dxgi.h>
 
 namespace mozilla {
 namespace gfx {
@@ -52,6 +46,7 @@ decltype(D3D11CreateDevice)* sD3D11CreateDeviceFn = nullptr;
 
 // It should only be used within CreateDirectCompositionDevice.
 decltype(DCompositionCreateDevice2)* sDcompCreateDevice2Fn = nullptr;
+decltype(DCompositionCreateDevice3)* sDcompCreateDevice3Fn = nullptr;
 
 // It should only be used within CreateDCompSurfaceHandle
 decltype(DCompositionCreateSurfaceHandle)* sDcompCreateSurfaceHandleFn =
@@ -72,14 +67,15 @@ DeviceManagerDx::DeviceManagerDx()
     : mDeviceLock("gfxWindowsPlatform.mDeviceLock"),
       mCompositorDeviceSupportsVideo(false) {
   // Set up the D3D11 feature levels we can ask for.
-  if (IsWin8OrLater()) {
-    mFeatureLevels.AppendElement(D3D_FEATURE_LEVEL_11_1);
-  }
+  mFeatureLevels.AppendElement(D3D_FEATURE_LEVEL_11_1);
   mFeatureLevels.AppendElement(D3D_FEATURE_LEVEL_11_0);
   mFeatureLevels.AppendElement(D3D_FEATURE_LEVEL_10_1);
   mFeatureLevels.AppendElement(D3D_FEATURE_LEVEL_10_0);
+  MOZ_COUNT_CTOR(DeviceManagerDx);
 }
 
+DeviceManagerDx::~DeviceManagerDx() { MOZ_COUNT_DTOR(DeviceManagerDx); }
+
 bool DeviceManagerDx::LoadD3D11() {
   FeatureState& d3d11 = gfxConfig::GetFeature(Feature::D3D11_COMPOSITING);
   MOZ_ASSERT(d3d11.IsEnabled());
@@ -112,12 +108,11 @@ bool DeviceManagerDx::LoadD3D11() {
 
 bool DeviceManagerDx::LoadDcomp() {
   MOZ_ASSERT(gfxConfig::GetFeature(Feature::D3D11_COMPOSITING).IsEnabled());
-  MOZ_ASSERT(gfxVars::UseWebRender());
   MOZ_ASSERT(gfxVars::UseWebRenderANGLE());
   MOZ_ASSERT(gfxVars::UseWebRenderDCompWin());
 
   if (sDcompCreateDevice2Fn) {
-    return true;
+    return true;  // Already loaded.
   }
 
   nsModuleHandle module(LoadLibrarySystem32(L"dcomp.dll"));
@@ -127,6 +122,8 @@ bool DeviceManagerDx::LoadDcomp() {
 
   sDcompCreateDevice2Fn = (decltype(DCompositionCreateDevice2)*)GetProcAddress(
       module, "DCompositionCreateDevice2");
+  sDcompCreateDevice3Fn = (decltype(DCompositionCreateDevice3)*)GetProcAddress(
+      module, "DCompositionCreateDevice3");
   if (!sDcompCreateDevice2Fn) {
     return false;
   }
@@ -135,9 +132,6 @@ bool DeviceManagerDx::LoadDcomp() {
   sDcompCreateSurfaceHandleFn =
       (decltype(DCompositionCreateSurfaceHandle)*)::GetProcAddress(
           module, "DCompositionCreateSurfaceHandle");
-  if (!sDcompCreateDevice2Fn) {
-    return false;
-  }
 
   mDcompModule.steal(module);
   return true;
@@ -272,6 +266,11 @@ static inline bool ProcessOwnsCompositor() {
 #endif
 
 bool DeviceManagerDx::CreateCompositorDevices() {
+  MutexAutoLock lock(mDeviceLock);
+  return CreateCompositorDevicesLocked();
+}
+
+bool DeviceManagerDx::CreateCompositorDevicesLocked() {
   MOZ_ASSERT(ProcessOwnsCompositor());
 
   FeatureState& d3d11 = gfxConfig::GetFeature(Feature::D3D11_COMPOSITING);
@@ -279,7 +278,7 @@ bool DeviceManagerDx::CreateCompositorDevices() {
 
   if (int32_t sleepSec =
           StaticPrefs::gfx_direct3d11_sleep_on_create_device_AtStartup()) {
-    printf_stderr("Attach to PID: %d\n", GetCurrentProcessId());
+    printf_stderr("Attach to PID: %lu\n", GetCurrentProcessId());
     Sleep(sleepSec * 1000);
   }
 
@@ -293,11 +292,6 @@ bool DeviceManagerDx::CreateCompositorDevices() {
     MOZ_ASSERT(!mCompositorDevice);
     ReleaseD3D11();
 
-    // Sync Advanced-Layers with D3D11.
-    if (gfxConfig::IsEnabled(Feature::ADVANCED_LAYERS)) {
-      gfxConfig::SetFailed(Feature::ADVANCED_LAYERS, FeatureStatus::Unavailable,
-                           "Requires D3D11", "FEATURE_FAILURE_NO_D3D11"_ns);
-    }
     return false;
   }
 
@@ -316,7 +310,7 @@ bool DeviceManagerDx::CreateCompositorDevices() {
   // Fallback from WR to D3D11 Non-WR compositor without re-creating gpu process
   // could happen when WR causes error. In this case, the attachments are loaded
   // synchronously.
-  if (!gfx::gfxVars::UseWebRender() || gfx::gfxVars::UseSoftwareWebRender()) {
+  if (gfx::gfxVars::UseSoftwareWebRender()) {
     PreloadAttachmentsOnCompositorThread();
   }
 
@@ -339,7 +333,7 @@ bool DeviceManagerDx::CreateVRDevice() {
     return false;
   }
 
-  RefPtr<IDXGIAdapter1> adapter = GetDXGIAdapter();
+  RefPtr<IDXGIAdapter1> adapter = GetDXGIAdapterLocked();
   if (!adapter) {
     NS_WARNING("Failed to acquire a DXGI adapter for VR");
     return false;
@@ -362,6 +356,11 @@ bool DeviceManagerDx::CreateVRDevice() {
 }
 
 bool DeviceManagerDx::CreateCanvasDevice() {
+  MutexAutoLock lock(mDeviceLock);
+  return CreateCanvasDeviceLocked();
+}
+
+bool DeviceManagerDx::CreateCanvasDeviceLocked() {
   MOZ_ASSERT(ProcessOwnsCompositor());
 
   if (mCanvasDevice) {
@@ -372,7 +371,7 @@ bool DeviceManagerDx::CreateCanvasDevice() {
     return false;
   }
 
-  RefPtr<IDXGIAdapter1> adapter = GetDXGIAdapter();
+  RefPtr<IDXGIAdapter1> adapter = GetDXGIAdapterLocked();
   if (!adapter) {
     NS_WARNING("Failed to acquire a DXGI adapter for Canvas");
     return false;
@@ -387,11 +386,28 @@ bool DeviceManagerDx::CreateCanvasDevice() {
     return false;
   }
 
+  if (StaticPrefs::
+          gfx_direct2d_target_independent_rasterization_disabled_AtStartup()) {
+    int creationFlags = 0x2;  // disable target independent rasterization
+    const GUID D2D_INTERNAL_DEVICE_CREATION_OPTIONS = {
+        0xfb3a8e1a,
+        0x2e3c,
+        0x4de1,
+        {0x84, 0x42, 0x40, 0x43, 0xe0, 0xb0, 0x94, 0x95}};
+    mCanvasDevice->SetPrivateData(D2D_INTERNAL_DEVICE_CREATION_OPTIONS,
+                                  sizeof(creationFlags), &creationFlags);
+  }
+
   if (FAILED(hr) || !mCanvasDevice) {
     NS_WARNING("Failed to acquire a D3D11 device for Canvas");
     return false;
   }
 
+  if (!D3D11Checks::DoesTextureSharingWork(mCanvasDevice)) {
+    mCanvasDevice = nullptr;
+    return false;
+  }
+
   if (XRE_IsGPUProcess()) {
     Factory::SetDirect3D11Device(mCanvasDevice);
   }
@@ -400,6 +416,11 @@ bool DeviceManagerDx::CreateCanvasDevice() {
 }
 
 void DeviceManagerDx::CreateDirectCompositionDevice() {
+  MutexAutoLock lock(mDeviceLock);
+  CreateDirectCompositionDeviceLocked();
+}
+
+void DeviceManagerDx::CreateDirectCompositionDeviceLocked() {
   if (!gfxVars::UseWebRenderDCompWin()) {
     return;
   }
@@ -421,10 +442,16 @@ void DeviceManagerDx::CreateDirectCompositionDevice() {
   HRESULT hr;
   RefPtr<IDCompositionDesktopDevice> desktopDevice;
   MOZ_SEH_TRY {
-    hr = sDcompCreateDevice2Fn(
+    hr = sDcompCreateDevice3Fn(
         dxgiDevice.get(),
         IID_PPV_ARGS(
             (IDCompositionDesktopDevice**)getter_AddRefs(desktopDevice)));
+    if (!desktopDevice) {
+      hr = sDcompCreateDevice2Fn(
+          dxgiDevice.get(),
+          IID_PPV_ARGS(
+              (IDCompositionDesktopDevice**)getter_AddRefs(desktopDevice)));
+    }
   }
   MOZ_SEH_EXCEPT(EXCEPTION_EXECUTE_HANDLER) { return; }
 
@@ -460,16 +487,26 @@ HANDLE DeviceManagerDx::CreateDCompSurfaceHandle() {
 void DeviceManagerDx::ImportDeviceInfo(const D3D11DeviceStatus& aDeviceStatus) {
   MOZ_ASSERT(!ProcessOwnsCompositor());
 
+  MutexAutoLock lock(mDeviceLock);
   mDeviceStatus = Some(aDeviceStatus);
 }
 
-void DeviceManagerDx::ExportDeviceInfo(D3D11DeviceStatus* aOut) {
+bool DeviceManagerDx::ExportDeviceInfo(D3D11DeviceStatus* aOut) {
+  MutexAutoLock lock(mDeviceLock);
   if (mDeviceStatus) {
     *aOut = mDeviceStatus.value();
+    return true;
   }
+
+  return false;
 }
 
 void DeviceManagerDx::CreateContentDevices() {
+  MutexAutoLock lock(mDeviceLock);
+  CreateContentDevicesLocked();
+}
+
+void DeviceManagerDx::CreateContentDevicesLocked() {
   MOZ_ASSERT(gfxConfig::IsEnabled(Feature::D3D11_COMPOSITING));
 
   if (!LoadD3D11()) {
@@ -485,7 +522,12 @@ void DeviceManagerDx::CreateContentDevices() {
   }
 }
 
-IDXGIAdapter1* DeviceManagerDx::GetDXGIAdapter() {
+already_AddRefed<IDXGIAdapter1> DeviceManagerDx::GetDXGIAdapter() {
+  MutexAutoLock lock(mDeviceLock);
+  return do_AddRef(GetDXGIAdapterLocked());
+}
+
+IDXGIAdapter1* DeviceManagerDx::GetDXGIAdapterLocked() {
   if (mAdapter) {
     return mAdapter;
   }
@@ -494,20 +536,37 @@ IDXGIAdapter1* DeviceManagerDx::GetDXGIAdapter() {
   decltype(CreateDXGIFactory1)* createDXGIFactory1 =
       (decltype(CreateDXGIFactory1)*)GetProcAddress(dxgiModule,
                                                     "CreateDXGIFactory1");
-
   if (!createDXGIFactory1) {
     return nullptr;
   }
+  static const auto fCreateDXGIFactory2 =
+      (decltype(CreateDXGIFactory2)*)GetProcAddress(dxgiModule,
+                                                    "CreateDXGIFactory2");
 
   // Try to use a DXGI 1.1 adapter in order to share resources
   // across processes.
   RefPtr<IDXGIFactory1> factory1;
-  HRESULT hr =
-      createDXGIFactory1(__uuidof(IDXGIFactory1), getter_AddRefs(factory1));
-  if (FAILED(hr) || !factory1) {
-    // This seems to happen with some people running the iZ3D driver.
-    // They won't get acceleration.
-    return nullptr;
+  if (StaticPrefs::gfx_direct3d11_enable_debug_layer_AtStartup()) {
+    RefPtr<IDXGIFactory2> factory2;
+    if (fCreateDXGIFactory2) {
+      auto hr = fCreateDXGIFactory2(DXGI_CREATE_FACTORY_DEBUG,
+                                    __uuidof(IDXGIFactory2),
+                                    getter_AddRefs(factory2));
+      MOZ_ALWAYS_TRUE(!FAILED(hr));
+    } else {
+      NS_WARNING(
+          "fCreateDXGIFactory2 not loaded, cannot create debug IDXGIFactory2.");
+    }
+    factory1 = factory2;
+  }
+  if (!factory1) {
+    HRESULT hr =
+        createDXGIFactory1(__uuidof(IDXGIFactory1), getter_AddRefs(factory1));
+    if (FAILED(hr) || !factory1) {
+      // This seems to happen with some people running the iZ3D driver.
+      // They won't get acceleration.
+      return nullptr;
+    }
   }
 
   if (!mDeviceStatus) {
@@ -632,7 +691,7 @@ void DeviceManagerDx::CreateCompositorDevice(FeatureState& d3d11) {
     return;
   }
 
-  RefPtr<IDXGIAdapter1> adapter = GetDXGIAdapter();
+  RefPtr<IDXGIAdapter1> adapter = GetDXGIAdapterLocked();
   if (!adapter) {
     d3d11.SetFailed(FeatureStatus::Unavailable,
                     "Failed to acquire a DXGI adapter",
@@ -683,15 +742,12 @@ void DeviceManagerDx::CreateCompositorDevice(FeatureState& d3d11) {
 
   uint32_t featureLevel = device->GetFeatureLevel();
   auto formatOptions = D3D11Checks::FormatOptions(device);
-  {
-    MutexAutoLock lock(mDeviceLock);
-    mCompositorDevice = device;
+  mCompositorDevice = device;
 
-    int32_t sequenceNumber = GetNextDeviceCounter();
-    mDeviceStatus = Some(D3D11DeviceStatus(
-        false, textureSharingWorks, featureLevel, DxgiAdapterDesc::From(desc),
-        sequenceNumber, formatOptions));
-  }
+  int32_t sequenceNumber = GetNextDeviceCounter();
+  mDeviceStatus = Some(D3D11DeviceStatus(
+      false, textureSharingWorks, featureLevel, DxgiAdapterDesc::From(desc),
+      sequenceNumber, formatOptions));
   mCompositorDevice->SetExceptionMode(0);
 }
 
@@ -774,12 +830,7 @@ void DeviceManagerDx::CreateWARPCompositorDevice() {
     return;
   }
 
-  // Only test for texture sharing on Windows 8 since it puts the device into
-  // an unusable state if used on Windows 7
-  bool textureSharingWorks = false;
-  if (IsWin8OrLater()) {
-    textureSharingWorks = D3D11Checks::DoesTextureSharingWork(device);
-  }
+  bool textureSharingWorks = D3D11Checks::DoesTextureSharingWork(device);
 
   DXGI_ADAPTER_DESC desc;
   D3D11Checks::GetDxgiDesc(device, &desc);
@@ -787,15 +838,12 @@ void DeviceManagerDx::CreateWARPCompositorDevice() {
   int featureLevel = device->GetFeatureLevel();
 
   auto formatOptions = D3D11Checks::FormatOptions(device);
-  {
-    MutexAutoLock lock(mDeviceLock);
-    mCompositorDevice = device;
+  mCompositorDevice = device;
 
-    int32_t sequenceNumber = GetNextDeviceCounter();
-    mDeviceStatus = Some(D3D11DeviceStatus(
-        true, textureSharingWorks, featureLevel, DxgiAdapterDesc::From(desc),
-        sequenceNumber, formatOptions));
-  }
+  int32_t sequenceNumber = GetNextDeviceCounter();
+  mDeviceStatus = Some(D3D11DeviceStatus(
+      true, textureSharingWorks, featureLevel, DxgiAdapterDesc::From(desc),
+      sequenceNumber, formatOptions));
   mCompositorDevice->SetExceptionMode(0);
 
   reporterWARP.SetSuccessful();
@@ -804,7 +852,7 @@ void DeviceManagerDx::CreateWARPCompositorDevice() {
 FeatureStatus DeviceManagerDx::CreateContentDevice() {
   RefPtr<IDXGIAdapter1> adapter;
   if (!mDeviceStatus->isWARP()) {
-    adapter = GetDXGIAdapter();
+    adapter = GetDXGIAdapterLocked();
     if (!adapter) {
       gfxCriticalNote << "Could not get a DXGI adapter";
       return FeatureStatus::Unavailable;
@@ -847,10 +895,7 @@ FeatureStatus DeviceManagerDx::CreateContentDevice() {
     MOZ_ASSERT(ok);
   }
 
-  {
-    MutexAutoLock lock(mDeviceLock);
-    mContentDevice = device;
-  }
+  mContentDevice = device;
   mContentDevice->SetExceptionMode(0);
 
   RefPtr<ID3D10Multithread> multi;
@@ -862,30 +907,27 @@ FeatureStatus DeviceManagerDx::CreateContentDevice() {
   return FeatureStatus::Available;
 }
 
-RefPtr<ID3D11Device> DeviceManagerDx::CreateDecoderDevice() {
-  bool isAMD = false;
-  {
-    MutexAutoLock lock(mDeviceLock);
-    if (!mDeviceStatus) {
-      return nullptr;
-    }
-    isAMD = mDeviceStatus->adapter().VendorId == 0x1002;
+RefPtr<ID3D11Device> DeviceManagerDx::CreateDecoderDevice(
+    bool aHardwareWebRender) {
+  MutexAutoLock lock(mDeviceLock);
+
+  if (!mDeviceStatus) {
+    return nullptr;
   }
 
+  bool isAMD = mDeviceStatus->adapter().VendorId == 0x1002;
   bool reuseDevice = false;
-  if (StaticPrefs::gfx_direct3d11_reuse_decoder_device() < 0) {
-    // Use the default logic, which is to allow reuse of devices on AMD, but
-    // create separate devices everywhere else.
-    if (isAMD) {
-      reuseDevice = true;
-    }
-  } else if (StaticPrefs::gfx_direct3d11_reuse_decoder_device() > 0) {
+  if (gfxVars::ReuseDecoderDevice()) {
     reuseDevice = true;
+  } else if (isAMD) {
+    reuseDevice = true;
+    gfxCriticalNoteOnce << "Always have to reuse decoder device on AMD";
   }
 
   if (reuseDevice) {
-    if (mCompositorDevice && mCompositorDeviceSupportsVideo &&
-        !mDecoderDevice) {
+    // Use mCompositorDevice for decoder device only for hardware WebRender.
+    if (aHardwareWebRender && mCompositorDevice &&
+        mCompositorDeviceSupportsVideo && !mDecoderDevice) {
       mDecoderDevice = mCompositorDevice;
 
       RefPtr<ID3D10Multithread> multi;
@@ -907,7 +949,7 @@ RefPtr<ID3D11Device> DeviceManagerDx::CreateDecoderDevice() {
     return nullptr;
   }
 
-  RefPtr<IDXGIAdapter1> adapter = GetDXGIAdapter();
+  RefPtr<IDXGIAdapter1> adapter = GetDXGIAdapterLocked();
   if (!adapter) {
     return nullptr;
   }
@@ -935,104 +977,55 @@ RefPtr<ID3D11Device> DeviceManagerDx::CreateDecoderDevice() {
   return device;
 }
 
-RefPtr<MLGDevice> DeviceManagerDx::GetMLGDevice() {
-  MutexAutoLock lock(mDeviceLock);
-  if (!mMLGDevice) {
-    MutexAutoUnlock unlock(mDeviceLock);
-    CreateMLGDevice();
-  }
-  return mMLGDevice;
+// ID3D11DeviceChild, IDXGIObject and ID3D11Device implement SetPrivateData with
+// the exact same parameters.
+template <typename T>
+static HRESULT SetDebugName(T* d3d11Object, const char* debugString) {
+  return d3d11Object->SetPrivateData(WKPDID_D3DDebugObjectName,
+                                     strlen(debugString), debugString);
 }
 
-static void DisableAdvancedLayers(FeatureStatus aStatus,
-                                  const nsCString aMessage,
-                                  const nsCString& aFailureId) {
-  if (!NS_IsMainThread()) {
-    NS_DispatchToMainThread(NS_NewRunnableFunction(
-        "DisableAdvancedLayers", [aStatus, aMessage, aFailureId]() -> void {
-          DisableAdvancedLayers(aStatus, aMessage, aFailureId);
-        }));
-    return;
-  }
-
-  MOZ_ASSERT(NS_IsMainThread());
-
-  FeatureState& al = gfxConfig::GetFeature(Feature::ADVANCED_LAYERS);
-  if (!al.IsEnabled()) {
-    return;
-  }
-
-  al.SetFailed(aStatus, aMessage.get(), aFailureId);
-
-  FeatureFailure info(aStatus, aMessage, aFailureId);
-  if (GPUParent* gpu = GPUParent::GetSingleton()) {
-    Unused << gpu->SendUpdateFeature(Feature::ADVANCED_LAYERS, info);
-  }
-
-  if (aFailureId.Length()) {
-    nsString failureId = NS_ConvertUTF8toUTF16(aFailureId.get());
-    Telemetry::ScalarAdd(Telemetry::ScalarID::GFX_ADVANCED_LAYERS_FAILURE_ID,
-                         failureId, 1);
-  }
-
-  // Notify TelemetryEnvironment.jsm.
-  if (RefPtr<nsIObserverService> obs =
-          mozilla::services::GetObserverService()) {
-    obs->NotifyObservers(nullptr, "gfx-features-ready", nullptr);
-  }
-}
-
-void DeviceManagerDx::CreateMLGDevice() {
-  MOZ_ASSERT(layers::CompositorThreadHolder::IsInCompositorThread());
-
-  RefPtr<ID3D11Device> d3d11Device = GetCompositorDevice();
-  if (!d3d11Device) {
-    DisableAdvancedLayers(FeatureStatus::Unavailable,
-                          "Advanced-layers requires a D3D11 device"_ns,
-                          "FEATURE_FAILURE_NEED_D3D11_DEVICE"_ns);
-    return;
-  }
-
-  RefPtr<MLGDeviceD3D11> device = new MLGDeviceD3D11(d3d11Device);
-  if (!device->Initialize()) {
-    DisableAdvancedLayers(FeatureStatus::Failed, device->GetFailureMessage(),
-                          device->GetFailureId());
-    return;
-  }
-
-  // While the lock was unheld, we should not have created an MLGDevice, since
-  // this should only be called on the compositor thread.
+RefPtr<ID3D11Device> DeviceManagerDx::CreateMediaEngineDevice() {
   MutexAutoLock lock(mDeviceLock);
-  MOZ_ASSERT(!mMLGDevice);
-
-  // Only set the MLGDevice if the compositor device is still the same.
-  // Otherwise we could possibly have a bad MLGDevice if a device reset
-  // just occurred.
-  if (mCompositorDevice == d3d11Device) {
-    mMLGDevice = device;
+  if (!LoadD3D11()) {
+    return nullptr;
   }
+
+  HRESULT hr;
+  RefPtr<ID3D11Device> device;
+  UINT flags = D3D11_CREATE_DEVICE_VIDEO_SUPPORT |
+               D3D11_CREATE_DEVICE_BGRA_SUPPORT |
+               D3D11_CREATE_DEVICE_PREVENT_INTERNAL_THREADING_OPTIMIZATIONS;
+  if (!CreateDevice(nullptr, D3D_DRIVER_TYPE_HARDWARE, flags, hr, device)) {
+    return nullptr;
+  }
+  if (FAILED(hr) || !device || !D3D11Checks::DoesDeviceWork()) {
+    return nullptr;
+  }
+  Unused << SetDebugName(device.get(), "MFMediaEngineDevice");
+
+  RefPtr<ID3D10Multithread> multi;
+  device->QueryInterface(__uuidof(ID3D10Multithread), getter_AddRefs(multi));
+  if (multi) {
+    multi->SetMultithreadProtected(TRUE);
+  }
+  return device;
 }
 
 void DeviceManagerDx::ResetDevices() {
-  // Flush the paint thread before revoking all these singletons. This
-  // should ensure that the paint thread doesn't start mixing and matching
-  // old and new objects together.
-  if (PaintThread::Get()) {
-    CompositorBridgeChild* cbc = CompositorBridgeChild::Get();
-    if (cbc) {
-      cbc->FlushAsyncPaints();
-    }
-  }
-
   MutexAutoLock lock(mDeviceLock);
+  ResetDevicesLocked();
+}
 
+void DeviceManagerDx::ResetDevicesLocked() {
   mAdapter = nullptr;
   mCompositorAttachments = nullptr;
-  mMLGDevice = nullptr;
   mCompositorDevice = nullptr;
   mContentDevice = nullptr;
   mCanvasDevice = nullptr;
   mImageDevice = nullptr;
+  mVRDevice = nullptr;
+  mDecoderDevice = nullptr;
   mDirectCompositionDevice = nullptr;
   mDeviceStatus = Nothing();
   mDeviceResetReason = Nothing();
@@ -1040,8 +1033,10 @@ void DeviceManagerDx::ResetDevices() {
 }
 
 bool DeviceManagerDx::MaybeResetAndReacquireDevices() {
+  MutexAutoLock lock(mDeviceLock);
+
   DeviceResetReason resetReason;
-  if (!HasDeviceReset(&resetReason)) {
+  if (!HasDeviceResetLocked(&resetReason)) {
     return false;
   }
 
@@ -1052,20 +1047,20 @@ bool DeviceManagerDx::MaybeResetAndReacquireDevices() {
   bool createCanvasDevice = !!mCanvasDevice;
   bool createDirectCompositionDevice = !!mDirectCompositionDevice;
 
-  ResetDevices();
+  ResetDevicesLocked();
 
-  if (createCompositorDevice && !CreateCompositorDevices()) {
+  if (createCompositorDevice && !CreateCompositorDevicesLocked()) {
     // Just stop, don't try anything more
     return true;
   }
   if (createContentDevice) {
-    CreateContentDevices();
+    CreateContentDevicesLocked();
   }
   if (createCanvasDevice) {
-    CreateCanvasDevice();
+    CreateCanvasDeviceLocked();
   }
   if (createDirectCompositionDevice) {
-    CreateDirectCompositionDevice();
+    CreateDirectCompositionDeviceLocked();
   }
 
   return true;
@@ -1115,7 +1110,10 @@ static DeviceResetReason HResultToResetReason(HRESULT hr) {
 
 bool DeviceManagerDx::HasDeviceReset(DeviceResetReason* aOutReason) {
   MutexAutoLock lock(mDeviceLock);
+  return HasDeviceResetLocked(aOutReason);
+}
 
+bool DeviceManagerDx::HasDeviceResetLocked(DeviceResetReason* aOutReason) {
   if (mDeviceResetReason) {
     if (aOutReason) {
       *aOutReason = mDeviceResetReason.value();
@@ -1150,10 +1148,6 @@ static inline bool DidDeviceReset(const RefPtr<ID3D11Device>& aDevice,
 }
 
 bool DeviceManagerDx::GetAnyDeviceRemovedReason(DeviceResetReason* aOutReason) {
-  // Caller must own the lock, since we access devices directly, and can be
-  // called from any thread.
-  mDeviceLock.AssertCurrentThreadOwns();
-
   if (DidDeviceReset(mCompositorDevice, aOutReason) ||
       DidDeviceReset(mContentDevice, aOutReason) ||
       DidDeviceReset(mCanvasDevice, aOutReason)) {
@@ -1250,6 +1244,7 @@ RefPtr<IDCompositionDevice2> DeviceManagerDx::GetDirectCompositionDevice() {
 }
 
 unsigned DeviceManagerDx::GetCompositorFeatureLevel() const {
+  MutexAutoLock lock(mDeviceLock);
   if (!mDeviceStatus) {
     return 0;
   }
@@ -1270,28 +1265,6 @@ bool DeviceManagerDx::CanInitializeKeyedMutexTextures() {
          gfxVars::AllowD3D11KeyedMutex();
 }
 
-bool DeviceManagerDx::HasCrashyInitData() {
-  MutexAutoLock lock(mDeviceLock);
-  if (!mDeviceStatus) {
-    return false;
-  }
-
-  return (mDeviceStatus->adapter().VendorId == 0x8086 && !IsWin10OrLater());
-}
-
-bool DeviceManagerDx::CheckRemotePresentSupport() {
-  MOZ_ASSERT(XRE_IsParentProcess());
-
-  RefPtr<IDXGIAdapter1> adapter = GetDXGIAdapter();
-  if (!adapter) {
-    return false;
-  }
-  if (!D3D11Checks::DoesRemotePresentWork(adapter)) {
-    return false;
-  }
-  return true;
-}
-
 bool DeviceManagerDx::IsWARP() {
   MutexAutoLock lock(mDeviceLock);
   if (!mDeviceStatus) {
@@ -1425,20 +1398,12 @@ void DeviceManagerDx::PreloadAttachmentsOnCompositorThread() {
     return;
   }
 
-  bool enableAL = gfxConfig::IsEnabled(Feature::ADVANCED_LAYERS) &&
-                  !gfx::gfxVars::UseSoftwareWebRender();
-
   RefPtr<Runnable> task = NS_NewRunnableFunction(
-      "DeviceManagerDx::PreloadAttachmentsOnCompositorThread",
-      [enableAL]() -> void {
+      "DeviceManagerDx::PreloadAttachmentsOnCompositorThread", []() -> void {
         if (DeviceManagerDx* dm = DeviceManagerDx::Get()) {
-          if (enableAL) {
-            dm->GetMLGDevice();
-          } else {
-            RefPtr<ID3D11Device> device;
-            RefPtr<layers::DeviceAttachmentsD3D11> attachments;
-            dm->GetCompositorDevices(&device, &attachments);
-          }
+          RefPtr<ID3D11Device> device;
+          RefPtr<layers::DeviceAttachmentsD3D11> attachments;
+          dm->GetCompositorDevices(&device, &attachments);
         }
       });
   CompositorThread()->Dispatch(task.forget());
