@@ -2,19 +2,21 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-use api::{AlphaType, PremultipliedColorF, YuvFormat, YuvColorSpace};
+use api::{AlphaType, PremultipliedColorF, YuvFormat, YuvRangedColorSpace};
 use api::units::*;
+use crate::composite::CompositeFeatures;
 use crate::segment::EdgeAaSegmentMask;
-use crate::spatial_tree::{SpatialTree, ROOT_SPATIAL_NODE_INDEX, SpatialNodeIndex};
+use crate::spatial_tree::{SpatialTree, SpatialNodeIndex};
 use crate::gpu_cache::{GpuCacheAddress, GpuDataRequest};
 use crate::internal_types::FastHashMap;
 use crate::prim_store::ClipData;
 use crate::render_task::RenderTaskAddress;
-use crate::renderer::ShaderColorMode;
+use crate::render_task_graph::RenderTaskId;
+use crate::renderer::{ShaderColorMode, GpuBufferAddress};
 use std::i32;
 use crate::util::{TransformedRectKind, MatrixHelpers};
-use crate::glyph_rasterizer::SubpixelDirection;
-use crate::util::pack_as_float;
+use glyph_rasterizer::SubpixelDirection;
+use crate::util::{ScaleOffset, pack_as_float};
 
 // Contains type that must exactly match the same structures declared in GLSL.
 
@@ -56,6 +58,16 @@ impl ZBufferIdGenerator {
     }
 }
 
+#[derive(Clone, Debug)]
+#[repr(C)]
+#[cfg_attr(feature = "capture", derive(Serialize))]
+#[cfg_attr(feature = "replay", derive(Deserialize))]
+pub struct CopyInstance {
+    pub src_rect: DeviceRect,
+    pub dst_rect: DeviceRect,
+    pub dst_texture_size: DeviceSize,
+}
+
 #[derive(Debug, Copy, Clone)]
 #[cfg_attr(feature = "capture", derive(Serialize))]
 #[cfg_attr(feature = "replay", derive(Deserialize))]
@@ -99,8 +111,7 @@ pub struct BlurInstance {
 #[cfg_attr(feature = "replay", derive(Deserialize))]
 pub struct ScalingInstance {
     pub target_rect: DeviceRect,
-    pub source_rect: DeviceIntRect,
-    pub source_layer: i32,
+    pub source_rect: DeviceRect,
 }
 
 #[derive(Clone, Debug)]
@@ -224,6 +235,7 @@ pub struct ClipMaskInstance {
 }
 
 // 16 bytes per instance should be enough for anyone!
+#[repr(C)]
 #[derive(Debug, Clone)]
 #[cfg_attr(feature = "capture", derive(Serialize))]
 #[cfg_attr(feature = "replay", derive(Deserialize))]
@@ -231,98 +243,152 @@ pub struct PrimitiveInstanceData {
     data: [i32; 4],
 }
 
+/// Specifies that an RGB CompositeInstance's UV coordinates are normalized.
+const UV_TYPE_NORMALIZED: u32 = 0;
+/// Specifies that an RGB CompositeInstance's UV coordinates are not normalized.
+const UV_TYPE_UNNORMALIZED: u32 = 1;
+
+/// A GPU-friendly representation of the `ScaleOffset` type
+#[derive(Clone, Debug)]
+#[repr(C)]
+pub struct CompositorTransform {
+    pub sx: f32,
+    pub sy: f32,
+    pub tx: f32,
+    pub ty: f32,
+}
+
+impl CompositorTransform {
+    pub fn identity() -> Self {
+        CompositorTransform {
+            sx: 1.0,
+            sy: 1.0,
+            tx: 0.0,
+            ty: 0.0,
+        }
+    }
+}
+
+impl From<ScaleOffset> for CompositorTransform {
+    fn from(scale_offset: ScaleOffset) -> Self {
+        CompositorTransform {
+            sx: scale_offset.scale.x,
+            sy: scale_offset.scale.y,
+            tx: scale_offset.offset.x,
+            ty: scale_offset.offset.y,
+        }
+    }
+}
+
 /// Vertex format for picture cache composite shader.
 /// When editing the members, update desc::COMPOSITE
 /// so its list of instance_attributes matches:
 #[derive(Clone, Debug)]
 #[repr(C)]
 pub struct CompositeInstance {
-    // Device space rectangle of surface
-    rect: DeviceRect,
-    // Device space clip rect for this surface
+    // Picture space destination rectangle of surface
+    rect: PictureRect,
+    // Device space destination clip rect for this surface
     clip_rect: DeviceRect,
     // Color for solid color tiles, white otherwise
     color: PremultipliedColorF,
 
     // Packed into a single vec4 (aParams)
-    z_id: f32,
+    _padding: f32,
     color_space_or_uv_type: f32, // YuvColorSpace for YUV;
                                  // UV coordinate space for RGB
     yuv_format: f32,            // YuvFormat
-    yuv_rescale: f32,
+    yuv_channel_bit_depth: f32,
 
     // UV rectangles (pixel space) for color / yuv texture planes
     uv_rects: [TexelRect; 3],
 
-    // Texture array layers for color / yuv texture planes
-    texture_layers: [f32; 3],
+    // A 2d scale + offset transform for the rect
+    transform: CompositorTransform,
 }
 
 impl CompositeInstance {
     pub fn new(
-        rect: DeviceRect,
+        rect: PictureRect,
         clip_rect: DeviceRect,
         color: PremultipliedColorF,
-        layer: f32,
-        z_id: ZBufferId,
+        transform: CompositorTransform,
     ) -> Self {
         let uv = TexelRect::new(0.0, 0.0, 1.0, 1.0);
         CompositeInstance {
             rect,
             clip_rect,
             color,
-            z_id: z_id.0 as f32,
-            color_space_or_uv_type: pack_as_float(0u32),
+            _padding: 0.0,
+            color_space_or_uv_type: pack_as_float(UV_TYPE_NORMALIZED),
             yuv_format: 0.0,
-            yuv_rescale: 0.0,
-            texture_layers: [layer, 0.0, 0.0],
+            yuv_channel_bit_depth: 0.0,
             uv_rects: [uv, uv, uv],
+            transform,
         }
     }
 
     pub fn new_rgb(
-        rect: DeviceRect,
+        rect: PictureRect,
         clip_rect: DeviceRect,
         color: PremultipliedColorF,
-        layer: f32,
-        z_id: ZBufferId,
         uv_rect: TexelRect,
+        transform: CompositorTransform,
     ) -> Self {
         CompositeInstance {
             rect,
             clip_rect,
             color,
-            z_id: z_id.0 as f32,
-            color_space_or_uv_type: pack_as_float(1u32),
+            _padding: 0.0,
+            color_space_or_uv_type: pack_as_float(UV_TYPE_UNNORMALIZED),
             yuv_format: 0.0,
-            yuv_rescale: 0.0,
-            texture_layers: [layer, 0.0, 0.0],
+            yuv_channel_bit_depth: 0.0,
             uv_rects: [uv_rect, uv_rect, uv_rect],
+            transform,
         }
     }
 
     pub fn new_yuv(
-        rect: DeviceRect,
+        rect: PictureRect,
         clip_rect: DeviceRect,
-        z_id: ZBufferId,
-        yuv_color_space: YuvColorSpace,
+        yuv_color_space: YuvRangedColorSpace,
         yuv_format: YuvFormat,
-        yuv_rescale: f32,
-        texture_layers: [f32; 3],
+        yuv_channel_bit_depth: u32,
         uv_rects: [TexelRect; 3],
+        transform: CompositorTransform,
     ) -> Self {
         CompositeInstance {
             rect,
             clip_rect,
             color: PremultipliedColorF::WHITE,
-            z_id: z_id.0 as f32,
+            _padding: 0.0,
             color_space_or_uv_type: pack_as_float(yuv_color_space as u32),
             yuv_format: pack_as_float(yuv_format as u32),
-            yuv_rescale,
-            texture_layers,
+            yuv_channel_bit_depth: pack_as_float(yuv_channel_bit_depth),
             uv_rects,
+            transform,
         }
     }
+
+    // Returns the CompositeFeatures that can be used to composite
+    // this RGB instance.
+    pub fn get_rgb_features(&self) -> CompositeFeatures {
+        let mut features = CompositeFeatures::empty();
+
+        // If the UV rect covers the entire texture then we can avoid UV clamping.
+        // We should try harder to determine this for unnormalized UVs too.
+        if self.color_space_or_uv_type == pack_as_float(UV_TYPE_NORMALIZED)
+            && self.uv_rects[0] == TexelRect::new(0.0, 0.0, 1.0, 1.0)
+        {
+            features |= CompositeFeatures::NO_UV_CLAMP;
+        }
+
+        if self.color == PremultipliedColorF::WHITE {
+            features |= CompositeFeatures::NO_COLOR_MODULATION
+        }
+
+        features
+    }
 }
 
 /// Vertex format for issuing colored quads.
@@ -458,7 +524,7 @@ impl GlyphInstance {
 
 pub struct SplitCompositeInstance {
     pub prim_header_index: PrimitiveHeaderIndex,
-    pub polygons_address: GpuCacheAddress,
+    pub polygons_address: i32,
     pub z: ZBufferId,
     pub render_task_address: RenderTaskAddress,
 }
@@ -468,7 +534,7 @@ impl From<SplitCompositeInstance> for PrimitiveInstanceData {
         PrimitiveInstanceData {
             data: [
                 instance.prim_header_index.0,
-                instance.polygons_address.as_int(),
+                instance.polygons_address,
                 instance.z.0,
                 instance.render_task_address.0 as i32,
             ],
@@ -476,13 +542,69 @@ impl From<SplitCompositeInstance> for PrimitiveInstanceData {
     }
 }
 
+#[derive(Copy, Clone)]
+#[cfg_attr(feature = "capture", derive(Serialize))]
+#[cfg_attr(feature = "replay", derive(Deserialize))]
+pub struct QuadInstance {
+    pub render_task_address: RenderTaskAddress,
+    pub prim_address: GpuBufferAddress,
+    pub z_id: ZBufferId,
+    pub transform_id: TransformPaletteId,
+    pub quad_flags: u8,
+    pub edge_flags: u8,
+    pub part_index: u8,
+    pub segment_index: u8,
+}
+
+impl From<QuadInstance> for PrimitiveInstanceData {
+    fn from(instance: QuadInstance) -> Self {
+        /*
+            [32 bits prim address]
+            [8 bits quad flags] [8 bits edge flags] [16 bits render task address]
+            [8 bits segment flags] [24 bits z_id]
+            [8 bits segment index] [24 bits xf_id]
+         */
+        PrimitiveInstanceData {
+            data: [
+                instance.prim_address.as_int(),
+                ((instance.quad_flags as i32) << 24) |
+                ((instance.edge_flags as i32) << 16) |
+                instance.render_task_address.0 as i32,
+                ((instance.part_index as i32) << 24) | instance.z_id.0,
+                ((instance.segment_index as i32) << 24) | instance.transform_id.0 as i32,
+            ],
+        }
+    }
+}
+
+#[cfg_attr(feature = "capture", derive(Serialize))]
+pub struct QuadSegment {
+    pub rect: LayoutRect,
+    pub task_id: RenderTaskId,
+}
+
+#[repr(C)]
+#[derive(Clone)]
+#[cfg_attr(feature = "capture", derive(Serialize))]
+#[cfg_attr(feature = "replay", derive(Deserialize))]
+pub struct MaskInstance {
+    pub prim: PrimitiveInstanceData,
+    pub clip_transform_id: TransformPaletteId,
+    pub clip_address: i32,
+    pub info: [i32; 2],
+}
+
+
 bitflags! {
+    // Note: This can use up to 12 bits due to how it will
+    // be packed in the instance data.
+
     /// Flags that define how the common brush shader
     /// code should process this instance.
     #[cfg_attr(feature = "capture", derive(Serialize))]
     #[cfg_attr(feature = "replay", derive(Deserialize))]
     #[derive(MallocSizeOf)]
-    pub struct BrushFlags: u8 {
+    pub struct BrushFlags: u16 {
         /// Apply perspective interpolation to UVs
         const PERSPECTIVE_INTERPOLATION = 1;
         /// Do interpolation relative to segment rect,
@@ -496,10 +618,19 @@ bitflags! {
         const SEGMENT_REPEAT_X_ROUND = 16;
         /// Vertically follow border-image-repeat: round.
         const SEGMENT_REPEAT_Y_ROUND = 32;
+        /// Whether to position the repetitions so that the middle tile
+        /// is horizontally centered.
+        const SEGMENT_REPEAT_X_CENTERED = 64;
+        /// Whether to position the repetitions so that the middle tile
+        /// is vertically centered.
+        const SEGMENT_REPEAT_Y_CENTERED = 128;
         /// Middle (fill) area of a border-image-repeat.
-        const SEGMENT_NINEPATCH_MIDDLE = 64;
+        const SEGMENT_NINEPATCH_MIDDLE = 256;
         /// The extra segment data is a texel rect.
-        const SEGMENT_TEXEL_RECT = 128;
+        const SEGMENT_TEXEL_RECT = 512;
+        /// Whether to force the anti-aliasing when the primitive
+        /// is axis-aligned.
+        const FORCE_AA = 1024;
     }
 }
 
@@ -522,8 +653,8 @@ impl From<BrushInstance> for PrimitiveInstanceData {
                 ((instance.render_task_address.0 as i32) << 16)
                 | instance.clip_task_address.0 as i32,
                 instance.segment_index
-                | ((instance.edge_flags.bits() as i32) << 16)
-                | ((instance.brush_flags.bits() as i32) << 24),
+                | ((instance.brush_flags.bits() as i32) << 16)
+                | ((instance.edge_flags.bits() as i32) << 28),
                 instance.resource_address,
             ]
         }
@@ -569,12 +700,20 @@ impl TransformPaletteId {
 
     /// Extract the transform kind from the id.
     pub fn transform_kind(&self) -> TransformedRectKind {
-        if (self.0 >> 24) == 0 {
+        if (self.0 >> 23) == 0 {
             TransformedRectKind::AxisAligned
         } else {
             TransformedRectKind::Complex
         }
     }
+
+    /// Override the kind of transform stored in this id. This can be useful in
+    /// cases where we don't want shaders to consider certain transforms axis-
+    /// aligned (i.e. perspective warp) even though we may still want to for the
+    /// general case.
+    pub fn override_transform_kind(&self, kind: TransformedRectKind) -> Self {
+        TransformPaletteId((self.0 & 0x7FFFFFu32) | ((kind as u32) << 23))
+    }
 }
 
 /// The GPU data payload for a transform palette entry.
@@ -630,11 +769,20 @@ pub struct TransformPalette {
 }
 
 impl TransformPalette {
-    pub fn new(count: usize) -> Self {
+    pub fn new(
+        count: usize,
+    ) -> Self {
         let _ = VECS_PER_TRANSFORM;
+
+        let mut transforms = Vec::with_capacity(count);
+        let mut metadata = Vec::with_capacity(count);
+
+        transforms.push(TransformData::invalid());
+        metadata.push(TransformMetadata::invalid());
+
         TransformPalette {
-            transforms: vec![TransformData::invalid(); count],
-            metadata: vec![TransformMetadata::invalid(); count],
+            transforms,
+            metadata,
             map: FastHashMap::default(),
         }
     }
@@ -643,30 +791,13 @@ impl TransformPalette {
         self.transforms
     }
 
-    pub fn set_world_transform(
-        &mut self,
-        index: SpatialNodeIndex,
-        transform: LayoutToWorldTransform,
-    ) {
-        register_transform(
-            &mut self.metadata,
-            &mut self.transforms,
-            index,
-            ROOT_SPATIAL_NODE_INDEX,
-            // We know the root picture space == world space
-            transform.with_destination::<PicturePixel>(),
-        );
-    }
-
     fn get_index(
         &mut self,
         child_index: SpatialNodeIndex,
         parent_index: SpatialNodeIndex,
         spatial_tree: &SpatialTree,
     ) -> usize {
-        if parent_index == ROOT_SPATIAL_NODE_INDEX {
-            child_index.0 as usize
-        } else if child_index == parent_index {
+        if child_index == parent_index {
             0
         } else {
             let key = RelativeTransformKey {
@@ -690,8 +821,6 @@ impl TransformPalette {
                     register_transform(
                         metadata,
                         transforms,
-                        child_index,
-                        parent_index,
                         transform,
                     )
                 })
@@ -716,7 +845,24 @@ impl TransformPalette {
         let transform_kind = self.metadata[index].transform_kind as u32;
         TransformPaletteId(
             (index as u32) |
-            (transform_kind << 24)
+            (transform_kind << 23)
+        )
+    }
+
+    pub fn get_custom(
+        &mut self,
+        transform: LayoutToPictureTransform,
+    ) -> TransformPaletteId {
+        let index = register_transform(
+            &mut self.metadata,
+            &mut self.transforms,
+            transform,
+        );
+
+        let transform_kind = self.metadata[index].transform_kind as u32;
+        TransformPaletteId(
+            (index as u32) |
+            (transform_kind << 23)
         )
     }
 }
@@ -750,8 +896,11 @@ pub enum UvRectKind {
 pub struct ImageSource {
     pub p0: DevicePoint,
     pub p1: DevicePoint,
-    pub texture_layer: f32,
-    pub user_data: [f32; 3],
+    // TODO: It appears that only glyphs make use of user_data (to store glyph offset
+    // and scale).
+    // Perhaps we should separate the two so we don't have to push an empty unused vec4
+    // for all image sources.
+    pub user_data: [f32; 4],
     pub uv_rect_kind: UvRectKind,
 }
 
@@ -765,12 +914,7 @@ impl ImageSource {
             self.p1.x,
             self.p1.y,
         ]);
-        request.push([
-            self.texture_layer,
-            self.user_data[0],
-            self.user_data[1],
-            self.user_data[2],
-        ]);
+        request.push(self.user_data);
 
         // If this is a polygon uv kind, then upload the four vertices.
         if let UvRectKind::Quad { top_left, top_right, bottom_left, bottom_right } = self.uv_rect_kind {
@@ -789,8 +933,6 @@ impl ImageSource {
 fn register_transform(
     metadatas: &mut Vec<TransformMetadata>,
     transforms: &mut Vec<TransformData>,
-    from_index: SpatialNodeIndex,
-    to_index: SpatialNodeIndex,
     transform: LayoutToPictureTransform,
 ) -> usize {
     // TODO: refactor the calling code to not even try
@@ -807,17 +949,11 @@ fn register_transform(
         inv_transform,
     };
 
-    if to_index == ROOT_SPATIAL_NODE_INDEX {
-        let index = from_index.0 as usize;
-        metadatas[index] = metadata;
-        transforms[index] = data;
-        index
-    } else {
-        let index = transforms.len();
-        metadatas.push(metadata);
-        transforms.push(data);
-        index
-    }
+    let index = transforms.len();
+    metadatas.push(metadata);
+    transforms.push(data);
+
+    index
 }
 
 pub fn get_shader_opacity(opacity: f32) -> i32 {
