@@ -11,6 +11,7 @@
 
 #include "js/CallArgs.h"
 #include "js/Class.h"
+#include "js/ColumnNumber.h"  // JS::LimitedColumnNumberZeroOrigin
 #include "js/GCAPI.h"
 #include "js/HeapAPI.h"
 #include "js/Object.h"           // JS::GetClass
@@ -55,9 +56,9 @@ extern JS_PUBLIC_API bool JS_NondeterministicGetWeakSetKeys(
     JSContext* cx, JS::HandleObject obj, JS::MutableHandleObject ret);
 
 // Raw JSScript* because this needs to be callable from a signal handler.
-extern JS_PUBLIC_API unsigned JS_PCToLineNumber(JSScript* script,
-                                                jsbytecode* pc,
-                                                unsigned* columnp = nullptr);
+extern JS_PUBLIC_API unsigned JS_PCToLineNumber(
+    JSScript* script, jsbytecode* pc,
+    JS::LimitedColumnNumberZeroOrigin* columnp = nullptr);
 
 /**
  * Determine whether the given object is backed by a DeadObjectProxy.
@@ -388,6 +389,11 @@ JS_PUBLIC_API JSFunction* NewFunctionByIdWithReserved(JSContext* cx,
                                                       unsigned nargs,
                                                       unsigned flags, jsid id);
 
+/**
+ * Get or set function's reserved slot value.
+ * `fun` should be a function created with `*WithReserved` API above.
+ * Such functions have 2 reserved slots, and `which` can be either 0 or 1.
+ */
 JS_PUBLIC_API const JS::Value& GetFunctionNativeReserved(JSObject* fun,
                                                          size_t which);
 
