@@ -23,7 +23,7 @@
 #include "shell/OSObject.h"
 #include "shell/StringUtils.h"
 #include "util/Text.h"
-#include "vm/JSAtom.h"
+#include "vm/JSAtomUtils.h"  // AtomizeString, PinAtom
 #include "vm/JSContext.h"
 #include "vm/StringType.h"
 
@@ -459,7 +459,7 @@ JSObject* ModuleLoader::loadAndParse(JSContext* cx, HandleString pathArg) {
     return module;
   }
 
-  UniqueChars filename = JS_EncodeStringToLatin1(cx, path);
+  UniqueChars filename = JS_EncodeStringToUTF8(cx, path);
   if (!filename) {
     return nullptr;
   }
